

/*******************************************************************/
/* This program translates formulae from bliksem format into otter */
/* format.                                                         */
/* At this moment it does not work for clauses.                    */
/*******************************************************************/

#define CLASS 
#include "basics.h"
#include "extfunc.h" 
#include "resolution.h"

#define SHOWSTEPS   0 

/****************************************************************/
/*               Error functions:                               */
/****************************************************************/


/* Since the program is non-interactive, there is no recovery after */
/* an error.                                                        */

static void closefiles ( )
{
   int i;

   for( i = freeincludefiles - 1; i >= 0; i -- ) 
   {
      if( includefiles [i]. rw == FILE_READ )
      {
         fclose ( includefiles [i]. handle ); 
         includefiles [i]. rw = FILE_NOTINUSE;
      }   
   }
   if( inputfile. rw == FILE_READ )
   {
      fclose( inputfile. handle );
      inputfile. rw = FILE_NOTINUSE; 
   }
}

/* print s, followed by 'in line %d', and exit.   */

void errorwithline ( s )
char *s;
{
   int i;

   printf( "%s ", s );
  
   for( i = freeincludefiles - 1; i >= 0; i -- )
   {
      if( includefiles [i]. rw == FILE_READ ) 
      {
         printf( "in line %d of file %s\n", 
                 includefiles [i]. linenum,
                 includefiles [i]. name );
      }
   }

   if( inputfile. rw == FILE_READ ) 
   {
      printf("in line %d of file %s\n",
         inputfile. linenum, inputfile. name ); 
   }  
   printf("\n"); 
   closefiles ();
   printstatistics( &outputfile ); 
   exit(0); 
}


/* print s, and exit.                      */

void errorwithoutline ( s )
char *s;
{
   printf("%s", s );
   printf("\n\n");
   closefiles ();
   printstatistics( &outputfile ); 
   exit(0); 
}



/******************************************************************/
/*      Functions for printing clauses:                           */
/******************************************************************/


void printclause( f, clause )
bliksemfile *f;
int clause;
{
   int rule;
   int c;

   printstring( f, "(");
   printint( f, clauses [ clause + CLAUSE_NUMBER ] );
   printstring( f, ", " );
   printint( f, clauses [ clause + CLAUSE_GENERATION ] );
   printstring( f, ") ");

#if 0
   printf("[%x %x  %x %x,", 
      clauses [ clause + CLAUSE_FILT1 ],
      clauses [ clause + CLAUSE_FILT2 ],
      clauses [ clause + CLAUSE_FILT3 ],
      clauses [ clause + CLAUSE_FILT4 ] ); 
   printf(" %x %x  %x %x]",
      clauses [ clause + CLAUSE_FILT5 ],
      clauses [ clause + CLAUSE_FILT6 ],
      clauses [ clause + CLAUSE_FILT7 ],
      clauses [ clause + CLAUSE_FILT8 ] ); 
#endif
   printstring( f, "{" ); 
   printstring( f, "C" );
      printint( f, clauses [ clause + CLAUSE_COMP ] ); 
   printstring( f, ",H" );
      printint( f, clauses [ clause + CLAUSE_NEXTHASH ] );
   printstring( f, ",D" );
      printint( f, clauses [ clause + CLAUSE_DEPTH ] ); 
   printstring( f, ",L" );
      printint( f, clauses [ clause + CLAUSE_LENGTH ] ); 
   printstring( f, ",V" );
      printint( f, clauses [ clause + CLAUSE_NRVARS ] ); 
   printstring( f, ",M" );
      printint( f, clauses [ clause + CLAUSE_END ] - 
                   clauses [ clause + CLAUSE_MAXLITS ] ); 
   printstring( f, "} " );

   switch( clauses [ clause + CLAUSE_RULE ] )
   {
   case RULE_BINARY:
      printstring( f, "R" ); 
      break; 
   case RULE_PARAMOD:
      printstring( f, "P" ); 
      break; 
   case RULE_HYPERRES:
      printstring( f, "H" ); 
      break; 
   case RULE_EQREFL:
      printstring( f, "E" );
      break; 
   case RULE_INITIAL:
      printstring( f, "I" ); 
      break;
   default:   
      exit(0); 
   }
   if( clauses [ clause + CLAUSE_LITERALS ] - 
       clauses [ clause + CLAUSE_PARENTS ] > 0 )
   {
      printstring( f, "(" ); 
      for( c = clauses [ clause + CLAUSE_PARENTS ]; 
           c < clauses [ clause + CLAUSE_LITERALS ]; 
           c ++ ) 
      {
         printint( f, clauses [ clauses [c] + CLAUSE_NUMBER ] );
         if( c + 1 == clauses [ clause + CLAUSE_LITERALS ] ) 
            printstring( f, ")" );
         else
            printstring( f, "," );
      }
   }

   if( clauses [ clause + CLAUSE_FACTORS ] > 0 ) 
   {
      printstring( f, "+");
      printint( f, clauses [ clause + CLAUSE_FACTORS ] );
      printstring( f, "f "); 
   }

   /* the literals: */

   printstring( f, " { " ); 
   for( c = clauses [ clause + CLAUSE_LITERALS ]; 
        c < clauses [ clause + CLAUSE_END ];
        c ++ )
   {
      printterm( f, clauses [c] );
      if( c + 1 != clauses [ clause + CLAUSE_END ] )
         printstring( f, ", " );
   }
   printstring( f, " }.\n" ); 
}


void printclauselist( f, clause )
bliksemfile *f;
int clause;
{
   while( clause != EMPTYCLAUSELIST )
   {
      printclause( f, clause );
      clause = clauses [ clause + CLAUSE_NEXTINLIST ];
   }
}


void printotter2( f, term )
bliksemfile *f;
int term;
{
   int operator;
   int arity;
   int i;

   operator = termspace [ term ];
   arity = arities [ operator ];

   if( operator < 0 ) exit(0);

   if( arity == 0 )
   {
      switch( operator )
      {
      case OP_DOT:
         printstring( f, "$nil " );
         return;
      case OP_FALSE:
         printstring( f, "$false " );
         return;
      case OP_TRUE:
         printstring( f, "$true " );
         return;
      }
   }

   if( arity == 1 )
   {
      switch( operator )
      {
      case OP_NOT:
         printstring( f, "-" );
         goto ar1; 
      case OP_UNARYMINUS:
         printstring( f, "-" );
         goto ar1;
      case OP_UNARYPLUS:
         printstring( f, "+" );
      ar1:
         printstringnnl( f, "( ");
         printotter2( f, term + 1 );
         printstring( f, ") " );
         return;
      }
   } 
       
   if( arity == 2 )
   {
      switch( operator )
      {
      case OP_DOT:
         printstring( f, "$cons" );
         goto ar2;
      case OP_AND:
         printstring( f, "&" );
         goto ar2;
      case OP_OR:
         printstring( f, "|" );
         goto ar2;
      case OP_IMPLIES:
         printstring( f, "->" );
         goto ar2;
      case OP_EQUIV:
         printstring( f, "<->" );
         goto ar2;

      case OP_FORALL:
      case OP_EXISTS:
         if( operator == OP_FORALL )
            printstring( f, "$Quantified( all, " );
         else
            printstring( f, "$Quantified( exists, " );
         printotter2( f, term + 1 );
         printstring( f, ", " );
         printotter2( f, termends [ term + 1 ] );
         printstring( f, ") ");
         return;
      case OP_EQUAL:
         printstring( f, "=" );
         goto ar2;
      case OP_REWRITES:
         printstring( f, "=" );
         goto ar2;
      case OP_LESS:
         printstring( f, "<" );
         goto ar2;
      case OP_GREATER:
         printstring( f, ">" );
         goto ar2; 
      case OP_EQUALEQUAL:
         printstring( f, "==" );
         goto ar2; 
      case OP_LESSEQUAL:
         printstring( f, "<=" );
         goto ar2;
      case OP_GREATEREQUAL:
         printstring( f, ">=" ); 
         goto ar2;
      case OP_GREATERGREATER:
         printstring( f, ">>" );
         goto ar2;
      case OP_LESSLESS:
         printstring( f, "<<" );
         goto ar2;
      case OP_GREATERGREATEREQUAL:
         printstring( f, ">>=" );
         goto ar2;
      case OP_LESSLESSEQUAL:
         printstring( f, "<<=" );
         goto ar2;
      case OP_GREATERGREATEREQUALEQUAL:
         printstring( f, ">>==" );
         goto ar2;
      case OP_LESSLESSEQUALEQUAL:
         printstring( f, "<<==" );
         goto ar2;
      case OP_IN:
         printstring( f, ":" );
         goto ar2;
      case OP_PLUS:
         printstring( f, "+" );
         goto ar2;
      case OP_MINUS:
         printstring( f, "-" );
         goto ar2;
      case OP_TIMES:
         printstring( f, "*" );
         goto ar2;
      case OP_DIVIDES:
         printstring( f, "/" );
         goto ar2;
      case OP_POW:
         printstring( f, "^" );
      ar2:
         printstringnnl( f, "( " );   /* No space here. */
         printotter2( f, term + 1 );
         printstring( f, ", ");
         printotter2( f, termends [ term + 1 ] );
         printstring( f, ") ");
         return; 
      }
   }

   if( operator < OP_NONPRIMITIVE )
      exit(0);

   printname( f, operators [ operator ]. name );

   if( arity == 0 )
   {
      printstring( f, " ");
      return;
   }
   printstringnnl( f, "( ");
   term = term + 1;

   for( i = 0; i < arity; i ++ )
   {
      printotter2( f, term );
      term = termends [ term ];
      if( i < arity - 1 )
         printstring( f, ", ");
      else
         printstring( f, ") ");
   }
   return; 
}



void printotter( f, form )
bliksemfile *f;
int form;
{
   printstring( f, "  ");
   printotter2( f, form );
   printstring( f, ".\n" );
}






/*************************************************************/
/*            Additional functions:                          */
/*************************************************************/


static int tryflag( term, s, x, minval, maxval )
int term; 
char *s;
int *x;
int minval;
int maxval;
{
   char *s1;

   s1 = operatorstring( termspace [ term + 1 ] );
   if( strcmp( s, s1 ) != 0 )
      return 0;

   s1 = operatorstring( termspace [ term + 2 ] );
   sscanf( s1, "%d", x );
   if( *x < minval || *x > maxval )
   {
      printstring( &outputfile, "value for flag is out of range:\n" );
      printterm( &outputfile, term );
      errorwithoutline( "" );
      return 0;
   }
   return 1;
}
      




/************************************************************/
/*                       main                               */
/************************************************************/



int main ()
{
   int i; 
   int startofinput, endofinput; 
   int tt;
   int startoftranslation, endoftranslation;
   char *s;
   int *translation;


   initinoutput( );

   freetermspace = 0; 
   freeclauses = 0; 

   (void) parser ( );

#if SHOWSTEPS
   printoperators( &outputfile );
   printstring( &outputfile,"\n\n"); 
#endif

   i = 0;
   while( i < freetermspace )
   {
      maketermends( i );
      i = termends [i];
   }

   /* Skip the options. We do not attempt to translate them. */    

   i = 0;
   s = operatorstring( termspace [i] );
   while( s != (char*) NULL && strcmp( s, "Set" ) == 0 )
   {
      i = termends [i];  
      s = operatorstring( termspace [i] ); 
   }


   startofinput = i;
   endofinput = freetermspace; 

#if SHOWSTEPS
   printtermlist( &outputfile, "To be translated:", startofinput, endofinput );
   printstring( &outputfile, "\n" ); 
#endif

   /* Type checking: All formulae should be first order.  */

   startoftranslation = endofinput;
   for( i = startofinput; i < endofinput; i = termends [i] )
   {
      tt = hastype( i, TYPE_FOFORM );
      if( ! tt ) 
      {
         printf("input is ill-typed\n\n");
         return 0;
      } 
      (void) copyterm( i ); 
   }
   endoftranslation = freetermspace; 

   printstring( &outputfile, "\n" );
   printstring( &outputfile, "% Created by blik2ott\n\n");

   printstring( &outputfile, "set( hyper_res ).\n");
   printstring( &outputfile, "% set( binary_res ).\n");
   printstring( &outputfile, "% set( para_into ).\n");
   printstring( &outputfile, "% set( para_from ).\n");
   printstring( &outputfile, "% set( order_eq ).\n");
   printstring( &outputfile, "% clear( para_from_right ).\n" );

   printstring( &outputfile, "%assign( max_weight, 50 ).\n" );

   printstring( &outputfile, "set(factor).\n\n");

   printstring( &outputfile, "formula_list(sos).\n\n" );
   printstring( &outputfile, "% all x ( x = x ).\n\n");

   for( i = startofinput; i < endofinput; i = termends [i] )
   {
      printotter( &outputfile, i );
   }
   printstring( &outputfile, "end_of_list(sos)." ); 
   printstring( &outputfile, "\n\n");
 
   return 0;
}




