

/**************************************************************/
/* File objects.c contains simple operators belonging to      */
/* the datatypes.                                             */
/**************************************************************/


#define CLASS extern
#include "basics.h"
#include "extfunc.h"
#include "resolution.h"

/************************************************************/
/*       Type name:                                         */
/************************************************************/


/* compute the hash value of a string. */

static int hashvalue ( s )
char *s;
{
   int hv;

   hv = 0;
   while( *s )
   {
      hv *= 113;
      hv += ( *s & 0X7F ); 
      hv = hv % MAXHASHVALUE;
      s ++ ; 
   }
   return hv;
}

/* convert a string into a name:  */ 

int makename ( s )
char *s;
{
   int ne;
   int hv;
   int pos;

   if( strlen( s ) >= MAXNAMELEN - 3 ) 
      errorwithoutline( "string too long" );

   hv = hashvalue ( s );
   pos = hashindex [ hv ];
   while( pos >= 0 )
   {
      if( strcmp( s, namespace + nametable [ pos ]. string ) == 0 )
      {
         return pos;   
      }
      pos = nametable [ pos ]. next; 
   }
   
   /* the name must be created. */

   ne = freenamespace;
   while( *s )
   {
      namespace [ freenamespace ++ ] = * ( s ++ );
      if( freenamespace + 3 >= MAXNAMESPACE ) 
         errorwithline( "out of name space" );
   }
   namespace [ freenamespace ++ ] = 0;
   nametable [ freenametable ]. string = ne;
   nametable [ freenametable ]. next = hashindex [ hv ];
   hashindex [ hv ] = freenametable;
   if( freenametable + 3 >= MAXNAMETABLE )
      errorwithline( "nametable is full" );
   return freenametable ++ ;
} 
     
 
/* make a name from a template.                           */
/* the name will have form sXXX, where XXX is an integer. */

int makefreshname( s )
char *s;
{
   int p;
   int x;
   int res;
   char newname [ MAXNAMELEN ];

   for( p = 0; s[p]; p ++ )
   {
      newname [p] = s[p];
      if( p > 20 )
         internalproblem( "absurd fresh name");
   }
    
   res = freenametable;
   x = 1;
   while( freenametable == res ) 
   {
      sprintf( newname + p, "%d", x );
      (void) makename( newname );    
      x ++ ; 
   }       
   return res; 
}
   

/*************************************************************/
/*       Type operator:                                      */
/*************************************************************/


/* Produce the string that defines a user operator.   */

char *operatorstring( f )
int f;
{
   int n;

   if( f < OP_NONPRIMITIVE )
      return (char*) NULL;
   else
   {
      n = operators [f]. name;
      return namespace + nametable [n]. string;
   }
}



/* make a user defined operator from an atom + arity.        */

int makeuseroperator( at, arity )
int at;
int arity;
{
   int i;

   for( i = OP_NONPRIMITIVE; i < freeoperators; i ++ ) 
   {
      if( operators [i]. name == at && arities [i] == arity )
      {
         return i;
      }
   }
   if( freeoperators >= MAXOPERATORS - 4 ) 
      errorwithline( "there are too many operators" );

   operators [i]. name = at;
   operators [i]. flags = FLAG_USER;
   arities [i] = arity;
   freeoperators ++ ; 
   return i;
}


/* produce a new operator with the specified flag. */ 

int makefreshoperator( flag, arity )
int flag;
int arity;
{
   int res; 
   int at;

   switch( flag )
   {
   case FLAG_SKOLEM:
      at = makefreshname( "skol" );
      break; 
   case FLAG_SUBFORM:
      at = makefreshname( "alpha" );
      break; 
   case FLAG_SPLITTING: 
      at = makefreshname( "comp" );
      break; 
   case FLAG_VARIABLE:
      at = makefreshname( "xx" );
      break;
   case FLAG_WORLD: 
      at = makefreshname( "w" );
      break;
   case FLAG_ACCESS: 
      at = makefreshname( "R" );
      break; 
   default:
      internalproblem( "problem" );
   }

   if( freeoperators >= MAXOPERATORS - 4 ) 
      errorwithoutline( "there are too many operators" );

   res = freeoperators;
   freeoperators ++ ; 
   operators [ res ]. name = at;
   operators [ res ]. flags = flag;
   arities [ res ] = arity;
   return res;
}



/* Return 1 if the name of the operator starts with a capital letter. */ 

int isprologvar( op )
int op;
{
   int n, c;

   if( op >= OP_NONPRIMITIVE ) 
   { 
      n = operators [ op ]. name;
      c = namespace [ nametable [n]. string ];
      if( c >= 'A' && c <= 'Z' )
         return 1;
      else
         return 0;
   } 
   else
      return 0; 
}


/* Return 1 is the operator is a name.   */

int isname( op )
int op;
{
   if( op >= OP_NONPRIMITIVE )
      return 1;
   else
      return 0;
}



/***********************************************************/
/*                 Type termspace:                         */
/***********************************************************/


/* the difference with type term is that the termends are not */
/* used.                                                      */


/* count the terms from n up to freetermspace.  */ 

int countterms( n )
int n;
{
   int nr, remain;
   int a;

   nr = 0; 
   while( n < freetermspace )
   {
      remain = 1;
      while( remain > 0 )
      {
         a = termspace [ n ++ ];
         if( a < 0 )
            remain -- ;
         else
            remain += arities [ a ] - 1;
      }
      nr ++ ; 
   }
   return nr;
}


/* compute the ends of the subterms of term.  */

void maketermends ( term, end )
int term; 
int end; 
{
   int x;
   int i, a;
   int arity;

   if( end < term ) 
      internalproblem( "maketermends" );

   for( x = end - 1; x >= term; x -- )
   {
      a = termspace [ x ];
      if( a < 0 )
         termends [ x ] = x + 1;
      else
      {
         arity = arities [ a ];
         i = x + 1;
         while( arity > 0 ) 
         {
            i = termends [ i ];
            arity -- ;
         }
         termends [ x ] = i;
         if( i > end ) 
            internalproblem( "maketermends" );
      }
   }
}


/* insert n integers at position termspace [t].  */

void inserttermspace( t, n )
int t;
int n;
{
   int i, j;
   
   if( freetermspace + n >= MAXTERMSPACE - 4 )
      errorwithoutline( "termspace is full" );

   j = freetermspace + n - 1; 
   for( i = freetermspace - 1; i >= t; i -- ) 
   {
      termspace [ j ] = termspace [ i ] ;
      j -- ;
   }
   freetermspace += n;
}


/* delete n integers at position termspace [t].  */

void deletetermspace( t, n )
int t;
int n;
{
   int i;

   i = t + n;
   if( i > freetermspace ) 
      internalproblem( "problem is deletetermspace!");
   while( i < freetermspace )
      termspace [ t ++ ] = termspace [ i ++ ];
   freetermspace -= n;
}


/* insert the operator f at position t.    */

void insertoperator( f, t )
int f;
int t;
{
   inserttermspace( t, 1 );
   termspace [ t ] = f;
}


/* reverse from termspace [t1] to termspace [t2].  */ 

static void reversetermspace( t1, t2 )
int t1;
int t2;
{
   int c;

   t2 -- ; 
   while( t1 < t2 )
   {
      c = termspace [ t1 ];
      termspace [ t1 ] = termspace [ t2 ];
      termspace [ t2 ] = c;
      t1 ++ ; t2 -- ;
   }
}




/***************************************************************/
/*       Type term:                                            */
/***************************************************************/



/* exchange the two arguments of binary term t.   */

void exchangearguments( t )
int t;
{
   int a1, a2, a3;

   a1 = t + 1;
   a2 = termends [a1];
   a3 = termends [a2]; 
   reversetermspace( a1, a2 );
   reversetermspace( a2, a3 );
   reversetermspace( a1, a3 );
   maketermends( a1, a3 );
}


/* make a fresh copy of term on termspace [ freetermspace ].   */ 

int copyterm( term )
int term;
{
   int res;
   int endofterm;

   endofterm = termends [ term ];
   res = freetermspace;

   if( freetermspace + ( endofterm - term ) >= MAXTERMSPACE - 10 ) 
      errorwithoutline( "termspace is full" );

   while( term < endofterm ) 
      termspace [ freetermspace ++ ] = termspace [ term ++ ];
   maketermends ( res, freetermspace ); 
   return res;
}
 

/* make a fresh copy of term on termspace [ freetermspace ], with    */
/* the subterm of term on position replaced by value.                */


int replaceposition( term, position, value )
int term;
int position;
int value; 
{
   int res; 
   int i;

   res = freetermspace;
   if( freetermspace + 
          ( termends [ term ] - term ) + 
          ( termends [ value ] - value ) - 
          ( termends [ position ] - position ) >= MAXTERMSPACE - 5 ) 
      errorwithoutline( "termspace is full" );

   for( i = term; i < position; i ++ )
      termspace [ freetermspace ++ ] = termspace [i];
   for( i = value; i < termends [ value ]; i ++ )
      termspace [ freetermspace ++ ] = termspace [i];
   for( i = termends [ position ]; i < termends [ term ]; i ++ ) 
      termspace [ freetermspace ++ ] = termspace [i];
   maketermends ( res, freetermspace );
   return res;
}



/* return 1 if term1 and term2 are equal. */ 

int termequal( term1, term2 )
int term1;
int term2;
{
   int endterm1;

   int a; 
#if 0 
   printstring( &outputfile, "termequal " );
   printterm( &outputfile, term1 );
   printstring( &outputfile, " = " ); 
   printterm( &outputfile, term2 );
   printstring( &outputfile, "?\n" ); 
#endif
  
   endterm1 = termends [ term1 ]; 
   while( term1 < endterm1 )
   {
      a = termspace [ term1 ++ ];
      if( a != termspace [ term2 ++ ] )
         return 0;
   }
   return 1;
}


 
/*****************************************************************/
/*   Type of first order formulae                                */
/*****************************************************************/


/* It considers the variable in OP_FORALL and OP_EXISTS as       */
/* TYPE_TERM.                                                    */
/* -1 is returned on error.                                      */
/* It produces the types of the subterms, given the type of      */
/* the complete term.                                            */

static int subtermtype( termtype, op, n )
int termtype; 
int op;
int n;
{
   if( n < 0 || n >= arities [op] )
      return -1;

   if( isname( op ))
      return TYPE_TERM;

   switch( op )
   {
      /* arguments TYPE_TERM, result TYPE_TERM.     */

   case OP_NIL:
   case OP_DOT:
   case OP_PLUS:
   case OP_MINUS:
   case OP_TIMES:
   case OP_DIVIDES:
   case OP_POW:
   case OP_UNARYMINUS:
   case OP_UNARYPLUS:
      if( termtype != TYPE_TERM )
         return -1;
      return TYPE_TERM;
      
      /* arguments TYPE_TERM, result TYPE_FORM.    */

   case OP_EQUAL:
   case OP_REWRITES:

   case OP_LESS:
   case OP_GREATER:
   case OP_EQUALEQUAL:
   case OP_LESSEQUAL:
   case OP_GREATEREQUAL:
   case OP_GREATERGREATER:
   case OP_LESSLESS:
   case OP_GREATERGREATEREQUAL:
   case OP_LESSLESSEQUAL:
   case OP_GREATERGREATEREQUALEQUAL:
   case OP_LESSLESSEQUALEQUAL:
   case OP_IN:
   
      if( termtype != TYPE_FORM )
         return -1;
      return TYPE_TERM;

      /* argument TYPE_FORM, result TYPE_FORM.    */

   case OP_FALSE:
   case OP_TRUE:
   case OP_NOT:
   case OP_AND:
   case OP_OR:
   case OP_IMPLIES:
   case OP_EQUIV:
   case OP_NECESSARY:
   case OP_POSSIBLE:
      if( termtype != TYPE_FORM )
         return -1;
      return TYPE_FORM;

   case OP_FORALL:
   case OP_EXISTS:
      if( termtype != TYPE_FORM )
         return -1; 
      if( n == 0 )
         return TYPE_TERM; 
      if( n == 1 )    
         return TYPE_FORM;
        
   default:
      return -1;
   }  
}



/* operator op is free in term. The arity of op should be zero. */
/* operators are always used to represent variables, since      */
/* variables are zero arity operators.                          */

static int typedvarfree( op, term, termtype )
int op;
int term;
int termtype;
{
   int f;
   int i, arity;
   int st;

#if 0
   printstring( &outputfile, "typedvarfree " );
   printname( &outputfile, operators [ op ]. name );
   printstring( &outputfile, " in " ); 
   printterm( &outputfile, term );
   printstring( &outputfile, "  " ); 
   printtype( &outputfile, termtype );    
   printstring( &outputfile, "? \n" ); 
#endif

   if( termtype != TYPE_FORM && termtype != TYPE_TERM )
      internalproblem( "problem with termtype" );
 
   f = termspace [ term ++ ];
   arity = arities [ f ];

   if( f == OP_FORALL || f == OP_EXISTS ) 
   {
      if( termspace [ term ] == op )
         return 0;
      else
         return typedvarfree( op, term + 1, termtype );
   } 

   if( f == op && termtype == TYPE_TERM )
      return 1;

   for( i = 0; i < arity; i ++ ) 
   {
      st = subtermtype( termtype, f, i ); 
      if( st == -1 ) internalproblem("this should not happen");
      if( typedvarfree( op, term, st ))
         return 1;
      term = termends [ term ];
   }
   return 0;
}


int fovarfree( op, term )
int op;
int term;
{
   if( op < OP_NONPRIMITIVE || arities [ op ] != 0 ) 
      internalproblem( "incorrect first order variable" );
   return typedvarfree( op, term, TYPE_FORM );
}


/****  both this and the following function are termporary.  ***/

/* Destructively replace all free occurrences of op1 by op2. */
/* Both op1 and op2 should be arity 0 operators.             */


static void destreplacevar( term, termtype, op1, op2 )
int term;
int termtype; 
int op1;
int op2;
{
   int f;
   int i, arity;
   int st;

#if 0
   printstring( &outputfile, "      destreplacemetavar " );
   printterm( &outputfile, term );
   printstring( &outputfile," ["); 
   printtype( &outputfile, termtype ); 
   printstring( &outputfile, "] "); 
   printname( &outputfile, operators [ op1 ]. name );
   printstring( &outputfile, " => " );
   printname( &outputfile, operators [ op2 ]. name );
   printstring( &outputfile, "\n");
#endif
   if( termtype != TYPE_FORM && termtype != TYPE_TERM ) 
      internalproblem( "termtype is wrong" );
   
   f = termspace [ term ];
   arity = arities [ f ];
   if( f == op1 && termtype == TYPE_TERM )
   {
      termspace [ term ] = op2;
      return;
   }
   term ++ ; 

   if( f == OP_FORALL || f == OP_EXISTS ) 
   {
      if( termspace [ term ] == op1 ) 
         return; 
      else
         destreplacevar( term + 1, termtype, op1, op2 );
   }
   else
   {
      for( i = 0; i < arity; i ++ ) 
      {
         st = subtermtype( termtype, f, i ); 
         destreplacevar( term, st, op1, op2 );
         term = termends [ term ];
      }
   }
}
   
 
/* make a copy and substitute term [ op := value ].             */
/* term may be destroyed by replacements. No value is returned. */ 
/* op should be a zero arity operator. term and value are       */
/* complete terms.                                              */

static void fosubst2( term, termtype, op, value )
int term;
int op;
int value;
{
   int f; 
   int i, arity;
   int op1, op2;
   int st;

#if 0
   printstring( &outputfile, "fosubst2 " );
   printterm( &outputfile, term );
   printstring( &outputfile, " ["); 
   printtype( &outputfile, termtype );
   printstring( &outputfile, "] " );
   printname( &outputfile, operators [op]. name );
   printstring( &outputfile, " := " );
   printterm( &outputfile, value );
   printstring( &outputfile, "\n" );
#endif 

   f = termspace [ term ++ ];
   arity = arities [ f ];
   if( freetermspace >= MAXTERMSPACE - 5 ) 
      errorwithoutline( "termspace is full" );
   
   if( f == op && termtype == TYPE_TERM )
   {
      (void) copyterm( value );    
      return; 
   }
   if( f == OP_EXISTS || f == OP_FORALL ) 
   {
      termspace [ freetermspace ++ ] = f;
      op1 = termspace [ term ++ ]; 
      if( op1 == op )
      {
         termspace [ freetermspace ++ ] = op1; 
         (void) copyterm( term );
         return; 
      }
      else
      {
         if( fovarfree( op1, value ) && fovarfree( op, term ))
         {
            op2 = makefreshoperator( FLAG_VARIABLE, 0 );
            termspace [ freetermspace ++ ] = op2; 
            destreplacevar( term, termtype, op1, op2 );
         }
         else
         {
            termspace [ freetermspace ++ ] = op1;
            destreplacevar( term, termtype, op1, op1 );  
         }
         fosubst2( term, termtype, op, value ); 
         return;  
      }
   }
   termspace [ freetermspace ++ ] = f;
   for( i = 0; i < arity; i ++ ) 
   {
      st = subtermtype( termtype, f, i ); 
      fosubst2( term, st, op, value );
      term = termends [ term ];
   }
   return;
}


/* make a first order substitution. term is destroyed by fosubst.  */
/* term [ op := value ] is created on termspace [ freetermspace ].  */
/* op should be a zero arity operator.                              */

int fosubst( term, op, value )
int term;
int op;
int value;
{
   int res;

   if( op < 0 || arities [op] != 0 ) 
      internalproblem( "problem in fosubst" );

   res = freetermspace;
   fosubst2( term, TYPE_FORM, op, value );
   maketermends ( res, freetermspace );  
#if 0
   printstring( &outputfile, "result is: ");
   printterm( &outputfile, res );
   printstring( &outputfile, "\n" );
#endif 

   termspace [ term ] = -1;   
   termends [ term ] = term;
      /* term has become unreliable, so it should be destroyed.     */
      /* It is better to crash spectacularly, if one has to crash.  */ 
   return res; 
}


/***********************************************************/
/*         Type Type.                                      */
/***********************************************************/

#if 0

void printtype( f, t )
bliksemfile *f;
int t;
{
   switch( t )
   {
   case TYPE_FORM:
      printstring( f, "FORM" ); 
      return; 
   case TYPE_TERM:
      printstring( f, "TERM" );
      return;
   case TYPE_LISTCLAUSE:
      printstring( f, "LISTCLAUSE" );
      return;

   case TYPE_MODFORM:
      printstring( f, "MODFORM" );
      return;
   case TYPE_FOFORM:
      printstring( f, "FOFORM" );
      return;
   case TYPE_CLAUSE:
      printstring( f, "CLAUSE" );
      return;

   default:
      internalproblem( "unknown type" ); 
   }
}

#endif

static int hasmodaloperator( term )
int term;
{
   int i, end;
   int f;

   end = termends [ term ];
   
   for( i = term; i < end; i ++ ) 
   {
      f = termspace [ i ];
      if( f == OP_NECESSARY || f == OP_POSSIBLE ) 
         return 1;
   }
   return 0;
}

  
static int ismodal( term )
int term;
{
   int i, end;
   int f;

   end = termends [ term ];
   
   for( i = term; i < end; i ++ ) 
   {
      f = termspace [ i ];
      if( f != OP_FALSE && f != OP_TRUE && f != OP_NOT && 
          f != OP_AND && f != OP_OR && f != OP_IMPLIES && 
          f != OP_EQUIV && f != OP_NECESSARY && f != OP_POSSIBLE ) 
      {
         if( !isname( f ) || arities [f] > 0 || isprologvar( f )) 
            return 0;
      }
   }
   return 1;
}


/* check that term does not contain OP_AND, and OP_EXISTS, OP_FALSE,  */
/* OP_TRUE, OP_EQUIV, OP_IMPLIES, and that OP_NOT occurs only in a    */
/* literal.                                                           */


static int isfoclause( term )
int term;
{
   int i, end;
   int f;

   end = termends [ term ];

   for( i = term; i < end; i ++ )
   {
      f = termspace [i];
      if( f == OP_AND || f == OP_EXISTS || f == OP_IMPLIES || 
          f == OP_FALSE || f == OP_TRUE || f == OP_EQUIV )
         return 0;
      if( f == OP_NOT )
      {
         f = termspace [ i + 1 ];
         if( f == OP_NOT || f == OP_FORALL || f == OP_OR ) 
            return 0;
      }
   }
   return 1; 
}


/* Return 1 if term has a free prologvar of type TERM.    */ 
/* A prologvar is a 0-arity operator, whose name starts   */
/* with a capital. We forbid this, because it causes      */
/* confusion. It should be handled existentially, but     */
/* people might think that it is universal.               */

static int hasfreeprologvar( termbase, term, termtype )
int termbase;
int term;
int termtype;
{
   int f, arity;
   int i;
   int st;    
  
#if 0
   printstring( &outputfile, "hasfreeprologvar " );
   printterm( &outputfile, termbase );
   printstring( &outputfile, "\n  " ); 
   printterm( &outputfile, term );
   printstring( &outputfile, " [" );  
   printtype( &outputfile, termtype );
   printstring( &outputfile, "] ?\n");
#endif 
 
   f = termspace [ term ] ;
   arity = arities [f];

   if( isname(f) && arity == 0 && termtype == TYPE_TERM )
   {
      if( isprologvar( f ))
      {
         /* So this f should be bound.  */
  
         while( termbase != term )
         {
            if( termspace [ termbase ] == OP_FORALL || 
                termspace [ termbase ] == OP_EXISTS )
            {
               if( termspace [ termbase + 1 ] == f )
                  return 0;
            }
            termbase ++ ;
            while( termends [ termbase ] <= term )
               termbase = termends [ termbase ];  
         }
         return 1; 
      }
   }
   term ++ ; 
   for( i = 0; i < arity; i ++ )
   {
      st = subtermtype( termtype, f, i );
      if ( hasfreeprologvar( termbase, term, st ))
         return 1;
      term = termends [ term ]; 
   }
   return 0;
}




/* Check a basetype. This is always done recursively:   */ 

static int canhavetype( term, wantedtype )
int term;
int wantedtype;
{
   int i, arity;
   int f;
   int st;

#if 0
   printstring( &outputfile, "canhavetype " );
   printterm( &outputfile, term );
   printstring( &outputfile, ", " );
   printtype( &outputfile, wantedtype );
   printstring( &outputfile, " ?\n");
#endif

   f = termspace [ term ];
   if( f < 0 ) 
      internalproblem( "unexpected internal variable" ); 
   arity = arities [ f ];

   switch( wantedtype )
   {
   case TYPE_FORM:
      if ( f == OP_FORALL || f == OP_EXISTS ) 
      {
         term ++ ; 
         f = termspace [ term ]; 
         if( f < 0 ) 
            internalproblem( "unexpected internal variable" );
         if( !isname( f ))
            return 0;
         if( arities [f] > 0 )
            return 0;
         term ++ ; 
         return canhavetype( term, TYPE_FORM );
      }
      term ++ ; 
      for( i = 0; i < arity; i ++ )
      {
         st = subtermtype( wantedtype, f, i );
         if( st == -1 ) 
            return 0;
         if( ! canhavetype( term, st ))
            return 0;
         term = termends [ term ];
      }
      return 1;
   case TYPE_TERM:
      term ++ ;
      for( i = 0; i < arity; i ++ )
      {
         st = subtermtype( wantedtype, f, i );
         if( st == -1 ) 
            return 0;
         if( ! canhavetype( term, st ))
            return 0;
         term = termends [ term ];
      }
      return 1;
   case TYPE_LISTCLAUSE:
      if( f == OP_NIL )
         return 1;

      if( f != OP_DOT ) 
         return 0;

      term ++ ;       
      if( termspace [ term ] == OP_NOT ) 
         term ++ ;
      f = termspace [ term ];
      if ( f == OP_FALSE || f == OP_TRUE || f == OP_NOT || f == OP_AND || 
           f == OP_OR || f == OP_IMPLIES || f == OP_EQUIV || 
           f == OP_FORALL || f == OP_EXISTS || f == OP_NECESSARY || 
           f == OP_POSSIBLE ) 
         return 0;

      if( ! canhavetype( term, TYPE_FORM ))
         return 0;
      
      return canhavetype( termends [ term ], TYPE_LISTCLAUSE );

   default:
      internalproblem( "unknown type" );
      return 0; 
   }
}


int hastype( term, wantedtype )
int term;
int wantedtype;
{
#if 0
   printstring( &outputfile, "hastype " );
   printterm( &outputfile, term );
   printstring( &outputfile, ", " );
   printtype( &outputfile, wantedtype );
   printstring( &outputfile, " ?\n" );
#endif

   switch( wantedtype )
   {
   case TYPE_MODFORM:
      if( ! canhavetype( term, TYPE_FORM ))
         return 0;
      if( ! ismodal( term ))
         return 0;
      return 1;
   case TYPE_FOFORM:
      if( ! canhavetype( term, TYPE_FORM ))
         return 0;
      if( hasmodaloperator( term ))
         return 0; 
      if( hasfreeprologvar( term, term, TYPE_FORM ))
         return 0; 
      return 1;     
   case TYPE_CLAUSE:
      if( termspace [ term ] == OP_NIL || termspace [ term ] == OP_DOT )
      {
         if( ! canhavetype( term, TYPE_LISTCLAUSE ))
            return 0; 
      }
      else
      {
         if( !canhavetype( term, TYPE_FORM ))
            return 0;
         if( hasmodaloperator( term ))
            return 0;
         if( hasfreeprologvar( term, term, TYPE_FORM ))
            return 0;
         if( ! isfoclause( term ))
            return 0; 
      }
      return 1;
   default:
      internalproblem( "unknown type" ); 
   }
   return 1; 
}



/***************************************************************/
/*          Memory Management.                                 */
/***************************************************************/


void printstatistics( f )
bliksemfile *f;
{
   int i;
   int checksum;

   printstring( f, "Memory use:\n\n" );
   printstring( f, "space for terms:        " ); 
   printint( f, freetermspace );
   printstring( f, "\n" ); 
   
   printstring( f, "space for clauses:      " );
   printint( f, freeclauses );
   printstring( f, "\n\n\n" ); 

   printstring( f, "clauses generated:      " );
   printint( f, nrgenerated );
   printstring( f, "\n" );

   printstring( f, "clauses kept:           " );
   printint( f, nrkept );
   printstring( f, "\n" );

   printstring( f, "clauses selected:       " );
   printint( f, nrselected );
   printstring( f, "\n" );

   printstring( f, "clauses deleted:        " );
   printint( f, nrdeleted );
   printstring( f, "\n" ); 

   printstring( f, "clauses inuse deleted:  " );
   printint( f, nrdeletedinuse );
   printstring( f, "\n\n" );

   printstring( f, "subsentry:          " );
   printint( f, subsentry );
   printstring( f, "\n" ); 

   printstring( f, "literals s-matched: " );
   printint( f, subspropmatch );
   printstring( f, "\n" );

   printstring( f, "literals matched:   " );
   printint( f, subslitmatch );
   printstring( f, "\n" );
 
   printstring( f, "full subsumption:   " );
   printint( f, subsfullsubs );
   printstring( f, "\n\n" );

   checksum = 0;  
   for( i = 0; i < freetermspace; i ++ )
   {
      checksum ^= termspace [i];
   } 

   for( i = 0; i < freeclauses; i ++ )
   {
      checksum ^= clauses [i];
   }
 
   printstring( f, "checksum:           ");
   printint( f, checksum ); 
   printstring( f, "\n"); 

   printstring( f, "\n\n" ); 

}


