

/******************************************************************/
/* compute complexity of term, using operatororder [] and         */
/* operatorassoc [].                                              */
/* Complexity of a variable is always 1.                          */
/******************************************************************/


static int termcompl( term )
int term;
{
   int i, f;
   int compl, dir;
   int arity;

   f = termspace [ term ];
   if( f < 0 )
      return 1;

   dir = operatorassoc [f];
   arity = arities [f];
   compl = operatororder [f];
   term ++ ;

   /* Left higher compl than right. Causes wish for right associativity. */
   if( dir < 0 )
   {
      for( i = 0; i < arity; i ++ )
      {
         compl += ( arity - i ) * termcompl( term );
         term = termends [ term ];
      }
      return compl;
   }      

   /* Right higher compl than left. Causes wish for left associativity.  */
   if( dir > 0 )
   {
      for( i = 0; i < arity; i ++ )
      {
         compl += ( i + 1 ) * termcompl( term );
         term = termends [ term ];
      }
      return compl;
   }

   /* Left and right equal compl. */
   for( i = 0; i < arity; i ++ )
   {
      compl += termcompl( term );
      term = termends [ term ];
   }
   return compl; 
}




/*****************************************************************/
/*    Full Knuth-Bendix order, using operatororder [] and        */
/*    operatorassoc [].                                          */
/*****************************************************************/

/* Produce a profile of how often each variable occurs. */

static void varmultiset( term, factor )
int term;
int factor;
{
   int i, op;
   int dir, arity;

#if 0
   printstring( &outputfile, "varmultiset: ");
   printterm( &outputfile, term );
   printf(" %d\n", factor);
#endif

   op = termspace [ term ];
   if( op < 0 )
   {
      for( i = 0; i < freevarprofile; i ++ )
      {
         if( varprofile [i]. var1 == op )
         {
            varprofile [i]. depthfreq += factor; 
            return; 
         }
      }
      varprofile [ freevarprofile ]. var1 = op;
      varprofile [ freevarprofile ]. depthfreq = factor;
      freevarprofile ++ ;
      return;
   } 
  
   dir = operatorassoc [ op ];
   arity = arities [ op ];
   term ++ ;
 
   /* Tends to rightassociativity. */ 
   if( dir < 0 )
   {
      for( i = 0; i < arity; i ++ )
      {
         varmultiset( term, ( arity - i ) * factor );
         term = termends [ term ];
      }
      return;
   }

   /* Tends to leftassociativity. */
   if( dir > 0 )
   {
      for( i = 0; i < arity; i ++ )
      {
         varmultiset( term, ( i + 1 ) * factor );
         term = termends [ term ];
      }
      return;
   }

   for( i = 0; i < arity; i ++ )
   {
      varmultiset( term, factor );
      term = termends [ term ];
   }
   return;
}


static int withinvarmultiset( term, factor )
int term;
int factor;
{
   int i, op;
   int dir, arity;

#if 0 
   printstring( &outputfile, "withinvarmultiset: ");
   printterm( &outputfile, term );
   printf("[%d] ", termspace [ term ] );
   printf("  %d\n", factor);
#endif

   op = termspace [ term ];
   if( op < 0 )
   {
      for( i = 0; i < freevarprofile; i ++ )
      {
         if( varprofile [i]. var1 == op )
         {
            varprofile [i]. depthfreq -= factor ;
            if( varprofile [i]. depthfreq >= 0 )
               return 1;
            return 0; 
         }
      }
      return 0;
   } 
  
   dir = operatorassoc [ op ];
   arity = arities [ op ];
   term ++ ;
 
   /* Tends to rightassociativity. */ 
   if( dir < 0 )
   {
      for( i = 0; i < arity; i ++ )
      {
         if( !withinvarmultiset( term, ( arity - i ) * factor ))
            return 0;
         term = termends [ term ];
      }
      return 1;
   }

   /* Tends to leftassociativity. */
   if( dir > 0 )
   {
      for( i = 0; i < arity; i ++ )
      {
         if( !withinvarmultiset( term, ( i + 1 ) * factor ))
            return 0;
         term = termends [ term ];
      }
      return 1;
   }

   for( i = 0; i < arity; i ++ )
   {
      if( !withinvarmultiset( term, factor ))
         return 0;
      term = termends [ term ];
   }
   return 1;
}


static int kbo( term1, term2 )
int term1;
int term2;
{
   int d1, d2;

#if 0
   printstring( &outputfile, "kbo( " );
   printterm( &outputfile, term1 );
   printstring( &outputfile, ", " );
   printterm( &outputfile, term2 );
   printstring( &outputfile, ")\n" );
#endif 

   d1 = termcompl( term1 );
   d2 = termcompl( term2 );
   if( d1 < d2 )
   {
      freevarprofile = 0;
      varmultiset( term2, 1 );
      if( withinvarmultiset( term1, 1 ))
         return COMP_LESS;
      return COMP_UNCOMPARABLE;
   }
   if( d1 > d2 )
   {
      freevarprofile = 0;
      varmultiset( term1, 1 );
      if( withinvarmultiset( term2, 1 ))
         return COMP_MORE;
      return COMP_UNCOMPARABLE;
   }
   
   freevarprofile = 0;
   varmultiset( term1, 1 );
   if( withinvarmultiset( term2, 1 ))
   {
      freevarprofile = 0;
      varmultiset( term2, 1 );
      if( withinvarmultiset( term1, 1 ))
         return lifttermord( term1, term2 );
   }
   return COMP_UNCOMPARABLE;
}


