
/***********************************************************/
/* file rules.c contains the derivation rules.             */
/* Bliksem 1.10                                            */
/***********************************************************/

#define CLASS extern
#include "basics.h"
#include "extfunc.h"
#include "resolution.h"


/* Jumped to when the empty clause is found.   */

static jmp_buf emptyclause;

#define MAXPARDEPTH     30 


/* a priori sort a clause, using litorderings [].  */

static void sortclause( clause )
{
   int start, end;
   int i, m1, m2;
   int l;
   int cmp;

   start = clauses [ clause + CLAUSE_LITERALS ];
   end = clauses [ clause + CLAUSE_END ];
   
   m1 = start;
   m2 = end;

   while( m1 < m2 )
   {
      for( i = m1 + 1; i < end; i ++ )
      {
         cmp = literalcompare( clauses [ m1 ], clauses [i] );
         if( cmp == COMP_LESS ) 
         {
            m1 ++ ; 
            goto m1notmax;
         }
         if( cmp == COMP_MORE ) 
         {
            l = clauses [i];
            clauses [i] = clauses [ m1 ];
            clauses [ m1 ] = l;
            m1 ++ ;
            goto m1notmax;
         }
      }
      /* m1 is maximal. */
     
      m2 -- ; 
      l = clauses [ m1 ];
      clauses [ m1 ] = clauses [ m2 ];
      clauses [ m2 ] = l;
   m1notmax:
      ;
   }
   clauses [ clause + CLAUSE_MAXLITS ] = m2;
}


/* Select in clause, dependent on the value of litselect.   */

static void selectinclause( clause )
{
   int start, end;
   int i, realmax;
   int l, max;
   int cmp;

   if( litselect == LITSEL_NONE )
      return;

   if( litselect == LITSEL_POS || litselect == LITSEL_NEG )
      start = clauses [ clause + CLAUSE_LITERALS ];
   else
      start = clauses [ clause + CLAUSE_MAXLITS ];
   end = clauses [ clause + CLAUSE_END ];

   max = start;
   realmax = 0;
 
   for( i = start + 1; i < end; i ++ )
   {
      if( litselect == LITSEL_POS || litselect == LITSEL_POSORD )
         cmp = posselect( clauses [ max ], clauses [i] ); 
      else
         cmp = negselect( clauses [ max ], clauses [i] );
      if( cmp == COMP_LESS )
      {
         realmax = 1;
         max = i;
      }
      if( cmp == COMP_MORE )
         realmax = 1;
   }
   if( realmax )
   {
      l = clauses [ max ];
      clauses [ max ] = clauses [ end - 1 ];
      clauses [ end - 1 ] = l;
      clauses [ clause + CLAUSE_MAXLITS ] = end - 1;
   }
}



#if 1

/* Check if a clause is well-sorted, and if the order behaves well on it. */

static void checksorted ( clause )
int clause;
{
   int i1, i2, i3;
   int start, end;
   int cmp;

   start = clauses [ clause + CLAUSE_LITERALS ];
   end = clauses [ clause + CLAUSE_END ];

   /* check for consistent behaviour of the ordering: */

   for( i1 = start; i1 < end; i1 ++ )
      for( i2 = start; i2 < end; i2 ++ )
      {
         cmp = literalcompare( clauses [i1], clauses [i2] );
         if( cmp == COMP_LESS )
            if( literalcompare( clauses [i2], clauses [i1] ) != COMP_MORE )
               goto consfailure;

         if( cmp == COMP_MORE ) 
            if( literalcompare( clauses [i2], clauses [i1] ) != COMP_LESS ) 
               goto consfailure;

         if( cmp == COMP_UNCOMPARABLE ) 
            if( literalcompare( clauses [i2], clauses[i1] ) != 
                COMP_UNCOMPARABLE )
               goto consfailure;
         if( cmp == COMP_EQUIV ) 
            if( literalcompare( clauses [i2], clauses[i1] ) != COMP_EQUIV ) 
               goto consfailure;
      }


   /* check for transitivity. */

   for( i1 = start; i1 < end; i1 ++ )
      for( i2 = start; i2 < end; i2 ++ )
         for( i3 = start; i3 < end; i3 ++ )
         {
            if( literalcompare( clauses [i1], clauses[i2] ) == COMP_LESS && 
                literalcompare( clauses [i2], clauses[i3] ) == COMP_LESS ) 
            {
               if( literalcompare( clauses [i1], clauses[i3] ) != COMP_LESS ) 
                  goto transfailure;
            }
            if( literalcompare( clauses [i1], clauses[i2] ) == COMP_LESS &&
                literalcompare( clauses [i2], clauses[i3] ) == COMP_EQUIV )
            {
               if( literalcompare( clauses [i1], clauses[i3] ) != COMP_LESS )
                  goto transfailure;
            }
            if( literalcompare( clauses [i1], clauses[i2] ) == COMP_EQUIV &&
                literalcompare( clauses [i2], clauses[i3] ) == COMP_LESS )
            {
               if( literalcompare( clauses [i1], clauses[i3] ) != COMP_LESS )
                  goto transfailure;
            }
            if( literalcompare( clauses [i1], clauses[i2] ) == COMP_EQUIV &&
                literalcompare( clauses [i2], clauses[i3] ) == COMP_EQUIV )
            {
               if( literalcompare( clauses [i1], clauses[i3] ) != COMP_EQUIV )
                  goto transfailure;
            }
         }

   /* Check that all non-maximal literals are really non-maximal. */
   /* Check that all maximal literals are really maximal.         */

   for( i1 = start; i1 < end; i1 ++ )
   {
      for( i2 = start; i2 < end; i2 ++ )
      {
         if( literalcompare( clauses [i1], clauses[i2] ) == COMP_LESS )
         {
            if( i1 >= clauses [ clause + CLAUSE_MAXLITS ] )
               goto maxfailure;
            goto litnotmax;
         }
      }
      if( i1 < clauses [ clause + CLAUSE_MAXLITS ] )
         goto nonmaxfailure;
   litnotmax:
      ;
   }    

   return;

consfailure:
   printstring( &outputfile, "Consistency failure: " ); 
   printclause( &outputfile, clause );
   exit(0);
transfailure:
   printstring( &outputfile, "Transitivity failure: " ); 
   printclause( &outputfile, clause );
   exit(0);
maxfailure:
   printstring( &outputfile, "Maximality failure: " ); 
   printclause( &outputfile, clause );
   exit(0);
nonmaxfailure:
   printstring( &outputfile, "Non-maximality failure: " ); 
   printclause( &outputfile, clause );
   exit(0);
}

#endif


/* Direct the positive and negative equalities in a clause.   */

static void sortequalities( clause )
int clause;
{
   int i;
   int ord;
   int lit, at;
 
   for( i = clauses [ clause + CLAUSE_LITERALS ]; 
        i < clauses [ clause + CLAUSE_END ];
        i ++ ) 
   {
      lit = clauses [i]; 
      if( termspace [ lit ] == OP_NOT ) 
         at = lit + 1;
      else
         at = lit;
      if( termspace [ at ] == OP_EQUAL || 
          termspace [ at ] == OP_REWRITES ) 
      {
         ord = termcompare( at + 1, termends [ at + 1 ] );
         switch( ord )
         {
         case COMP_LESS: 
            termspace [ at ] = OP_REWRITES;
            exchangearguments( at ); 
            break;
         case COMP_MORE:
            termspace [ at ] = OP_REWRITES;
            break;
         default:
            termspace [ at ] = OP_EQUAL;
            break;
         }
      }
   }
}


/* Return 1 if term originating from origin, after the substitution,   */
/* is not smaller than the results of the maximal literals of clause   */
/* in literals []. In order to know where these results start,         */
/* aposteriori needs to know where the results of clause start.        */
/* This is indicated by firstlit.                                      */


static int aposteriori( firstlit, clause, term, origin )
int firstlit;
int clause;
int term;
int origin;
{
   int shouldbemax;
   int i;

   shouldbemax = ressubst( term, origin );
   maketermends( shouldbemax, freetermspace );

   i = firstlit + 
      ( clauses [ clause + CLAUSE_MAXLITS ] - 
        clauses [ clause + CLAUSE_LITERALS ] );
   while( i < nrliterals ) 
   {
      if( literalcompare( shouldbemax, literals [i] ) == COMP_LESS ) 
      {
         freetermspace = shouldbemax;
         return 0;
      }
      i ++ ; 
   }
   freetermspace = shouldbemax;
   return 1;
}

   
/* firstterm is the first term.                         */
/* No aposteriori check is made here.                   */

static void makenewclause( firstterm )
int firstterm; 
{
   int x;
   int i;
   int g, generation;
   int excusable;
   int result;

   result = freeclauses;

   if( showgenerated )
   {
      printstring( &outputfile, "Generated: (");
      printint( &outputfile, nrkept );
      printstring( &outputfile, ") { ");
      for( i = 0; i < nrliterals; i ++ )
      {
         printterm( &outputfile, literals [i] );
         if( i + 1 < nrliterals )
            printstring( &outputfile, ", " );
      }
      printstring( &outputfile, " }\n" );
   }
   nrgenerated ++ ;

   /* 123. Try to simplify the clause.            */
   /*    Check that it is not a tautology.        */
   /*    Remove identical literals + negative     */
   /*    equalities of the form ! t = t.          */

   if( literalsistautology( ))
      goto dontkeep; 

   simpfacteqrefl( firstterm );

   if( simpdemod( firstterm, -1, 1 ))
   {
      if( literalsistautology() )
         goto dontkeep;
      simpfacteqrefl( firstterm );
   }

   if( simpreso( firstterm, -1, 1 ))
   {
      if( literalsistautology( ) )
         goto dontkeep;
      simpfacteqrefl( firstterm );
   }

   /* 4. Really make the clause.                  */

   if( freeclauses + CLAUSE_REST + freeparents + 
       nrliterals >= MAXCLAUSES - 10 ) 
      errorwithoutline( "clauses is full" );
  
   freeclauses += CLAUSE_REST;
 
   clauses [ result + CLAUSE_NUMBER ] = nrkept;

   generation = -1;
   for( i = 0; i < freeparents; i ++ )
   {
      if( parents [i] >= 0 )
      {
         g = clauses [ parents [i] + CLAUSE_GENERATION ];
         if( g > generation )
            generation = g;
      }
   }
   clauses [ result + CLAUSE_GENERATION ] = generation + 1; 
   clauses [ result + CLAUSE_DERIVATION ] = freeclauses;
   for( i = 0; i < freeparents; i ++ )
      clauses [ freeclauses ++ ] = parents [i];
   clauses [ result + CLAUSE_LITERALS ] = freeclauses; 
   clauses [ result + CLAUSE_MAXLITS ] = freeclauses;

   for( i = 0; i < nrliterals; i ++ )
      clauses [ freeclauses ++ ] = literals [i];
 
   clauses [ result + CLAUSE_END ] = freeclauses;
   if( clauses [ result + CLAUSE_GENERATION ] <= excuselevel ) 
      excusable = 1;
   else
      excusable = 0;

   x = nrliterals;
   if( x > maxlength && !excusable ) 
   {
      incomplete = 1;
      goto dontkeep;
   }
   clauses [ result + CLAUSE_LENGTH ] = x;

   x = clauseweight( result );
   if( x > maxweight && !excusable ) 
   {
      incomplete = 1;
      goto dontkeep;
   }
   clauses [ result + CLAUSE_WEIGHT ] = x;

   x = clausedepth( result ); 
   if( x > maxdepth && !excusable ) 
   {
      incomplete = 1;
      goto dontkeep;
   }
   clauses [ result + CLAUSE_DEPTH ] = x;

   x = clausenrvars( result ); 
   if( x > maxnrvars && !excusable ) 
   {
      incomplete = 1;
      goto dontkeep; 
   }
   clauses [ result + CLAUSE_NRVARS ] = x; 

   makefilter( result );

   /* 5. forward subsumption check. */

   if( forwardsubsumed( result ))
      goto dontkeep; 

   /* Het is een blijvertje.        */

   nrkept ++ ; 

   clauses [ result + CLAUSE_NEXTINLIST ] = listunfactored;
   listunfactored = result; 

   /* 6. Check if we have the empty clause.    */

   if( clauses [ result + CLAUSE_LENGTH ] == 0 )
   {
      if( showkept )
      {
         printstring( &outputfile, "Kept:  " );
         printclause( &outputfile, result );
      }
      longjmp( emptyclause, 1 );
   }

   /* 7. Sort the clause, and direct the equalities, if  */
   /*    the proper options are set.                     */

   if( litapriori )
   {
      sortclause( result );
      /* checksorted( result );  */
      selectinclause( result );
   }
   if( termapriori )
      sortequalities( result );
   
   /* 8. Show the clause, if showkept = 1.               */
  
   if( showkept )
   {
      printstring( &outputfile, "Kept:  " );
      printclause( &outputfile, result );
   }

   /* 9. Add the most unlikely literal to the discrimination tree */
   /*    for forward subsumption.                                 */

   dt_subs( result, 1 );

   /* 10. Check if one can extract a demodulation rule from the   */
   /*     new clause.                                             */

   dt_demod( result, 1 );

   /* 11. Check if one can extract a resolution simplifier from   */
   /*     the new clause.                                         */

   dt_simpres( result, 1 );
 
   resimpinusecnt ++ ;
   resimpclausescnt ++ ;
   showstatuscnt ++ ;

   return;
dontkeep:
   freeclauses = result;
   freetermspace = firstterm;
   return;
}


/********************************************************/
/* processdeletedclause processes a deleted clause.     */
/* It does not do the actual deletion from listinuse,   */
/* or listclauses.                                      */
/* It prints the clause, (if showdeleted is on), and    */
/* removes it from all discrimination trees.            */
/* If frominuse =1, the clause was deleted from         */
/* list inuse. In that case different processing may be */
/* necessary.                                           */
/********************************************************/

#define DEL_BSUBS   100
#define DEL_SIMP    200
#define DEL_TOOMANY 300
   /* Possible reasons for deletion.   */

static void deleteclause( clause, frominuse, reason )
int *clause;
int frominuse;
int reason;
{
   int deleted;

   nrdeleted ++ ; 

   if( frominuse )
      nrdeletedinuse ++ ;

   deleted = *clause;
   *clause = clauses [ *clause + CLAUSE_NEXTINLIST ];

   if( showdeleted )
   { 
      switch( reason )
      {
      case DEL_BSUBS:
         printstring( &outputfile, "Backsubsumed: " );
         break;
      case DEL_SIMP:
         printstring( &outputfile, "Simplified: " );
         break;
      case DEL_TOOMANY:
         printstring( &outputfile, "Heaviest:   " );
         break;
      default:
         internalproblem( "deletion reason wrong" );
         exit(0);
      }
      printclause( &outputfile, deleted );
   } 

   if( frominuse )
   {
      clauses [ deleted + CLAUSE_NEXTINLIST ] = listdeletedinuse;
      listdeletedinuse = deleted;
   }
   else
   {
      clauses [ deleted + CLAUSE_NEXTINLIST ] = listdeleted;
      listdeleted = deleted;
   }

   dt_demod( deleted, 0 );
   dt_simpres( deleted, 0 );
   dt_subs( deleted, 0 );
}


/* Try to resimplify *clause. If it is possible then *clause is       */
/* replaced by its tail, and resimpclause returns 1.                  */
/* *clause should be a well-formed list of clauses.                   */


static int resimpclause( clause, frominuse )
int *clause;
int frominuse;
{
   int i;
   int firstterm;

#if 0
   if( frominuse )
      printstring( &outputfile, "Resimplifying clause from inuse:: ");
   else
      printstring( &outputfile, "Resimplifying clause: ");
   printclause( &outputfile, *clause );
#endif

   firstterm = freetermspace;

   if( backwardsubsumed( *clause ))
   {
      deleteclause( clause, frominuse, DEL_BSUBS ); 
      return 1;
   }  

   nrliterals = 0;
   for( i = clauses [ *clause + CLAUSE_LITERALS ]; 
        i < clauses [ *clause + CLAUSE_END ];
        i ++ )
   {
      literals [ nrliterals ++ ] = clauses [i];
   }

   parents [0] = RULE_RESIMP;
   parents [1] = *clause;
   freeparents = 2;

   if( simpdemod( 0, *clause, 0 ))
   {
      deleteclause( clause, frominuse, DEL_SIMP );
      makenewclause( firstterm );
      return 1;
   }

   if( simpreso( 0, *clause, 0 ))
   {
      deleteclause( clause, frominuse, DEL_SIMP );
      makenewclause( firstterm );
      return 1;
   }
   return 0;
}


/***********************************************************************/
/* Process a clause that is to become an axiom. It should be unsorted, */
/* and the equalities should be undirected. That's all.                */
/***********************************************************************/

static void processaxiomclause( clause )
int clause;
{
   int i;
   int lit;

#if 1
   /* Make all  literals maximal:   */

   clauses [ clause + CLAUSE_MAXLITS ] = 
      clauses [ clause + CLAUSE_LITERALS ];

   /* Undirect rewrites:            */
#endif
#if 0
   for( i = clauses [ clause + CLAUSE_LITERALS ];
        i < clauses [ clause + CLAUSE_END ];
        i ++ )
   {
      lit = clauses [i];
      if( termspace [ lit ] == OP_NOT ) 
         lit ++ ;
      if( termspace [ lit ] == OP_REWRITES ) 
         termspace [ lit ] = OP_EQUAL;
   }
#endif
   return;
}



/*****************************************************************/
/*   Produce an initial clause from a literal list.              */
/*   makeclause assumes very strongly that the                   */
/*   literal list has been typechecked already.                  */
/*****************************************************************/


/* translate a clause that it represented by a list. */ 

static void makeinit( term )
int term;
{
   nrliterals = 0;
   while( termspace [ term ] != OP_NIL ) 
   {
      term ++ ; 
      literals [ nrliterals ++ ] = copyterm( term );
      term = termends [ term ];

      if( nrliterals >= MAXLITERALS - 3 ) 
         errorwithoutline( "too many literals in clause" );
   }
}

   
/* Make an initial clause from term. It is assumed that t is   */
/* already typechecked by hastype( term, TYPE_CLAUSE ).        */

static void makeinitialclause ( term )
int term;
{
    int firstterm;

#if 0
   printstring( &outputfile, "Initial: " ); 
   printterm( &outputfile, term );
   printstring( &outputfile, "\n");
#endif
   firstterm = freetermspace;

   parents [0] = RULE_INITIAL;
   freeparents = 1;

   makeinit( term );
 
   makenewclause( firstterm );
}


/*******************************************************************/
/*                 Compute Factors:                                */ 
/*******************************************************************/

static void makefactors( clause )
int clause;
{
   int c1, c2;
   int firstterm;
   int i;

   for( c1 = clauses [ clause + CLAUSE_LITERALS ]; 
        c1 < clauses [ clause + CLAUSE_END ]; 
        c1 ++ )
   {
      for( c2 = clauses [ clause + CLAUSE_LITERALS ]; c2 < c1; c2 ++ )
      {
         freesubst = 0;
         if( unify( clauses [ c1 ], 0, clauses [ c2 ], 0 ))
         {
            firstterm = freetermspace;

            freerenaming = 0;     
            nrliterals = 0;
            for( i = clauses [ clause + CLAUSE_LITERALS ]; 
                 i < clauses [ clause + CLAUSE_END ];
                 i ++ )
            {
               if( i != c1 )
               {
                  literals [ nrliterals ++ ] = ressubst( clauses [i], 0 );
               }
            }
            parents [0] = RULE_FACTOR;
            parents [1] = clause;
            freeparents = 2;
            maketermends( firstterm, freetermspace );
            makenewclause( firstterm );         
         }
      }
   }
}



/*****************************************************************/
/*    Apply the equality reflexivity rule.                       */
/*****************************************************************/


static void makeeqrefls( clause )
int clause;
{
   int c;
   int op;
   int arg1, arg2; 
   int firstterm;
   int i;

   for( c = clauses [ clause + CLAUSE_LITERALS ]; 
        c < clauses [ clause + CLAUSE_END ]; 
        c ++ )
   {
      if( termspace [ clauses [c] ] == OP_NOT ) 
      {
         op = termspace [ clauses [c] + 1 ];
         if( op == OP_EQUAL || op == OP_REWRITES )
         {
            arg1 = clauses [c] + 2;
            arg2 = termends [ arg1 ];

            freesubst = 0;
            if( unify( arg1, 0, arg2, 0 ))
            {
               firstterm = freetermspace;
       
               freerenaming = 0;
               nrliterals = 0;
               for( i = clauses [ clause + CLAUSE_LITERALS ]; 
                    i < clauses [ clause + CLAUSE_END ];
                    i ++ )
               {
                  if( i != c )
                  {
                     literals [ nrliterals ++ ] = ressubst( clauses [i], 0 );
                  }
               }
               parents [0] = RULE_EQREFL;
               parents [1] = clause;
               freeparents = 2;
               maketermends( firstterm, freetermspace );
               makenewclause( firstterm );
            }
         }
      }
   }
}



/********************************************************************/
/* make equality factors of clause.                                 */
/********************************************************************/


static void makeeqfacts( clause )
int clause;
{
   int c1, c2, i;
   int lit1, lit2;
   int dir1, dir2;
   int t1, t2, t3, t1a;
   int t1s, t2s, t3s;
   int firstterm;
   int sideliterals;

   for( c1 = clauses [ clause + CLAUSE_MAXLITS ];
        c1 < clauses [ clause + CLAUSE_END ];
        c1 ++ )
   {
      lit1 = clauses [ c1 ];
      if( termspace [ lit1 ] == OP_EQUAL || 
          termspace [ lit1 ] == OP_REWRITES )
      {
         for( c2 = clauses [ clause + CLAUSE_LITERALS ];
              c2 < clauses [ clause + CLAUSE_END ];
              c2 ++ )
         {
            if( c1 != c2 )
            {
               lit2 = clauses [ c2 ];
               if( termspace [ lit2 ] == OP_EQUAL || 
                   termspace [ lit2 ] == OP_REWRITES ) 
               {
                  for( dir1 = 0; dir1 < 2; dir1 ++ )
                  {
                     if( termspace [ lit1 ] == OP_EQUAL ||
                         dir1 < 1 )
                     {
                        for( dir2 = 0; dir2 < 2; dir2 ++ )
                        {
                           if( termspace [ lit2 ] == OP_EQUAL ||
                               dir2 < 1 )
                           { 
                              if( dir1 == 0 )
                              { 
                                 t1 = lit1 + 1;
                                 t2 = termends [ t1 ];
                              }
                              else
                              {
                                 t2 = lit1 + 1;
                                 t1 = termends [ t2 ];
                              }
                      
                              if( dir2 == 0 )
                              { 
                                 t1a = lit2 + 1;
                                 t3 = termends [ t1a ];
                              }
                              else
                              {
                                 t3 = lit2 + 1;
                                 t1a = termends [ t3 ];
                              }

                              freesubst = 0;
                              if( unify( t1, 0, t1a, 0 ))
                              {
                                 firstterm = freetermspace;
                 
                                 freerenaming = 0;

                                 /* We now have:          */
                                 /* c1:   t1 = t2,        */
                                 /* c2:   t1 = t3.        */ 

                                 /* We must produce:      */
                                 /* ! t2 = t3, t1 = t3.   */

                                 literals [0] = freetermspace;
                                 termspace [ freetermspace ++ ] = OP_NOT;
                                 termspace [ freetermspace ++ ] = OP_EQUAL;
                                 t2s = ressubst( t2, 0 );
                                 t3s = ressubst( t3, 0 );
                                 t1s = ressubst( t1, 0 );

                                 maketermends( firstterm, freetermspace );

                                 if( termaposteriori &&
                                     termspace [ lit1 ] == OP_EQUAL &&
                                     termcompare( t1s, t2s ) == COMP_LESS ) 
                                 {
                                    freetermspace = firstterm;
                                    return;
                                 } 

                                 if( termaposteriori && 
                                     termspace [ lit2 ] == OP_EQUAL && 
                                     termcompare( t1s, t3s ) == COMP_LESS )
                                 {
                                    freetermspace = firstterm;
                                    return;
                                 }

                                 freetermspace = t1s;
                                 nrliterals = 1;
                                 sideliterals = freetermspace;

                                 for( 
                                    i = clauses [ clause + CLAUSE_LITERALS ];
                                    i < clauses [ clause + CLAUSE_END ];
                                    i ++ )
                                 {
                                    if( i != c1 )
                                    {
                                       literals [ nrliterals ++ ] = 
                                          ressubst( clauses [i], 0 );
                                    }
                                 }
                                
                                 maketermends( sideliterals, freetermspace ); 
                                 if( litaposteriori && 
                                     ! aposteriori( 1, clause, 
                                         clauses [ c1 ], 0 ))
                                 {
                                    freetermspace = firstterm;
                                    return;
                                 }
                                 parents [0] = RULE_EQFACT;
                                 parents [1] = clause;
                                 freeparents = 2;
                                 makenewclause( firstterm );
                              }
                           }
                        }
                     }
                  }
               }            
            }
         }  
      }
   }
   return;
}
 

/***************************************************/
/*     Binary Resolution:                          */
/***************************************************/

static int bin_clause;     /* reference in clauses.  */
static int bin_literal;    /* reference in clauses.  */
static int bin_term;       /* reference in literals. */


static void makebinary( clause, literal, dontcare )
int clause;
int literal;
int dontcare;
{
   int i;
   int literals2, termspace2;
   int firstterm;

   firstterm = freetermspace;

   freerenaming = 0;
   nrliterals = 0;

#if 0
   printstring( &outputfile, "makebinary: ");
   printclause( &outputfile, clause );
   printclause( &outputfile, bin_clause );
   printstring( &outputfile, "\n\n" );
#endif

   /* copy side literals from given: */

   for( i = clauses [ bin_clause + CLAUSE_LITERALS ];
        i < clauses [ bin_clause + CLAUSE_END ];
        i ++ ) 
   {
      if( i != bin_literal )
      {
         literals [ nrliterals ++ ] = ressubst( clauses [i], 0 );
      }
   }
   maketermends( firstterm, freetermspace ); 
 
   if( litaposteriori && 
       ! aposteriori( 0, bin_clause, clauses [ bin_literal ], 0 ))
   {
      freetermspace = firstterm;
      return;
   }

   literals2 = nrliterals;
   termspace2 = freetermspace;
 
   /* copy side literals from other: */

   for( i = clauses [ clause + CLAUSE_LITERALS ];
        i < clauses [ clause + CLAUSE_END ]; 
        i ++ ) 
   {
      if( i != literal )
      {
         literals [ nrliterals ++ ] = ressubst( clauses [i], 1 ); 
      }
   } 
   maketermends( termspace2, freetermspace );

   if( litaposteriori &&
       ! aposteriori( literals2, clause, clauses [ literal ], 1 ))
   {
      freetermspace = firstterm;
      return;
   }

   parents [0] = RULE_RES;
   parents [1] = bin_clause;
   parents [2] = clause;
   freeparents = 3;

   makenewclause( firstterm );
}


/* Look for possible binary resolvents.   */


static void searchbinary( given )
int given;
{
   int other;
   int givenfirst, givenlast;
   int otherfirst, otherlast; 
   int lit1, lit2;
   int c;

   bin_clause = given; 
   givenfirst = clauses [ bin_clause + CLAUSE_MAXLITS ];
   givenlast = clauses [ bin_clause + CLAUSE_END ];

   for( other = listinuse; other != EMPTYCLAUSELIST; 
        other = clauses [ other + CLAUSE_NEXTINLIST ] )
   {
      otherfirst = clauses [ other + CLAUSE_MAXLITS ]; 
      otherlast = clauses [ other + CLAUSE_END ];

      for( bin_literal = givenfirst; bin_literal < givenlast; 
           bin_literal ++ )
      {
         for( c = otherfirst; c < otherlast; c ++ ) 
         {
            lit1 = clauses [ bin_literal ];
            lit2 = clauses [ c ];

            if( termspace [ lit1 ] != OP_NOT && 
                termspace [ lit2 ] == OP_NOT && 
                termspace [ lit1 ] == termspace [ lit2 + 1 ] )
            {
               freesubst = 0; 
               if( unify( lit1, 0, lit2 + 1, 1 ))
                  goto resolvent;
            }
            if( termspace [ lit1 ] == OP_NOT && 
                termspace [ lit2 ] != OP_NOT && 
                termspace [ lit1 + 1 ] == termspace [ lit2 ] )
            {
               freesubst = 0;
               if( unify( lit1 + 1, 0, lit2, 1 ))
                  goto resolvent;
            }
            goto next; 
                                   
         resolvent:
            makebinary( other, c, 0 );
         next:
            ;
         }
      }
   }
   return;
}




/********************************************************************/
/*                  Paramodulation:                                 */
/********************************************************************/

/* The depth in literal of position.          */

static int positiondepth( literal, position )
int literal;
int position;
{
   int depth;

   if( position < literal || termends [ literal ] <= position )
      internalproblem( "position should be in literal" );

   if( termspace [ literal ] == OP_NOT ) 
      literal ++ ;

   depth = 0;
   while( literal != position )
   {
      literal ++ ;
      while( termends [ literal ] <= position )
         literal = termends [ literal ];
      depth ++ ;
   }
   return depth;
} 



/* replliterals( was, becomes ) replaces was by becomes everywhere */
/* in literals [].                                                 */

static void replliterals( was, becomes )
int was;
int becomes;
{
   int i;
   int t, tend;

   for( i = 0; i < nrliterals; i ++ )
   {
      t = literals [i];
      literals [i] = freetermspace;

      tend = termends [t];
      if( termspace [t] == OP_NOT ) 
         termspace [ freetermspace ++ ] = termspace [ t ++ ];
      termspace [ freetermspace ++ ] = termspace [ t ++ ];

      while( t < tend )
      {
         if( termequal( t, was ))
         {
            (void) copyterm( becomes ); 
            t = termends [t];
         }
         else
         {
            termspace [ freetermspace ++ ] = termspace [ t ++ ];
            if( freetermspace >= MAXTERMSPACE - 10 )
               errorwithoutline( "termspace is full" );
         }
      }
   }
}


static void moveliterals( from, to )
int from;
int to;
{
   int diff;
   int i;
   int firstterm;
 
   if( to > from ) 
   {
      internalproblem( "to > from" );
      exit(0);
   }

   diff = from - to;
   firstterm = to;

   while( from < freetermspace )
      termspace [ to ++ ] = termspace [ from ++ ];
   freetermspace = to;
   for( i = 0; i < nrliterals; i ++ )
      literals [i] -= diff;
   maketermends( firstterm, freetermspace ); 
}


/* make a paramodulant. The substitution has been constructed.     */
/* fromclause has index 0, and intoclause has index 1.             */
 

static void makepara( fromclause, fromlit, from1,
                     intoclause, intolit, intoterm )
int fromclause, fromlit, from1;
int intoclause, intolit, intoterm;
{
   int firstterm, termspace2;
   int i;
   int j1, j2;
   int from2;
   int from1subst, from2subst;
   int literals2;

#if 0
   printstring( &outputfile, "\n\nConstructing paramodulant:\n");
   printstring( &outputfile, "fromclause: " );
   printclause( &outputfile, fromclause );
   printterm( &outputfile, clauses [ fromlit ] );
   printstring( &outputfile, " [ ");
   printterm( &outputfile, from1 );
   printstring( &outputfile, " ]\n" );

   printstring( &outputfile, "intoclause: " );
   printclause( &outputfile, intoclause );
   printterm( &outputfile, clauses [ intolit ] );
   printstring( &outputfile, " [ ");
   printterm( &outputfile, intoterm );
   printstring( &outputfile, " ]\n" );
#endif

   if( positiondepth( clauses [ intolit ], intoterm ) > MAXPARDEPTH )
      return;
   freerenaming = 0;

   firstterm = freetermspace;
   freerenaming = 0;
   nrliterals = 0;

   if( clauses [ fromlit ] + 1 == from1 )
      from2 = termends [ from1 ]; 
   else
      from2 = clauses [ fromlit ] + 1;   

   /* construct equality literal after the substitution:    */

   from1subst = ressubst( from1, 0 ); 
   from2subst = ressubst( from2, 0 );
   maketermends( from1subst, freetermspace ); 

   /* Check if equality is directed wrong a posteriori:   */

   if( termaposteriori && termspace [ clauses [ fromlit ]] == OP_EQUAL 
       &&
       termcompare( from1subst, from2subst ) == COMP_LESS ) 
   {
      freetermspace = firstterm;
      return;
   }

   /* produce the complete into-clause under the substitution:  */

   for( i = clauses [ intoclause + CLAUSE_LITERALS ]; 
        i < clauses [ intoclause + CLAUSE_END ]; i ++ )
   {
      literals [ nrliterals ++ ] = ressubst( clauses [i], 1 );
   }
   maketermends( firstterm, freetermspace );

   /* Check the a posteriori maximality of the into-literal.  */

   if( litaposteriori && 
       ! aposteriori( 0, intoclause, clauses [ intolit ], 1 ))
   {
      freetermspace = firstterm; 
      return;
   }

   /* Now replace all occurrences of from1subst by from2subst in the */
   /* intoclause after the substitution.                             */

   j2 = freetermspace;
   replliterals( from1subst, from2subst );
   j1 = firstterm;
   moveliterals( j2, j1 );

   termspace2 = freetermspace;
   literals2 = nrliterals;

   /* copy side literals from from, using the substitution:  */

   for( i = clauses [ fromclause + CLAUSE_LITERALS ];
        i < clauses [ fromclause + CLAUSE_END ]; 
        i ++ ) 
   {
      if( i != fromlit )
      {
         literals [ nrliterals ++ ] = ressubst( clauses [i], 0 ); 
      }
   }
   maketermends( termspace2, freetermspace );

   if( litaposteriori && 
       ! aposteriori( literals2, fromclause, clauses [ fromlit ], 0 ))
   {
      freetermspace = firstterm;
      return;
   }

   parents [0] = RULE_PARAMOD;
   parents [1] = fromclause;
   parents [2] = intoclause;
   freeparents = 3;

   makenewclause( firstterm );
}


/* Both functions have the right form, so that they can  */
/* be called by retrievefromtree( ) in unify.c           */
/* However they are called by searchparamod.             */


/* paramodulate from bin_clause. */

static void makeparafrom( clause, literal, term )
{
   makepara( bin_clause, bin_literal, bin_term, clause, literal, term );
}


/* paramodulate into bin_clause. */

static void makeparainto( clause, literal, term )
{
   makepara( clause, literal, term, bin_clause, bin_literal, bin_term );
}



/* Produce possible paramodulants:             */

static void searchparamod( given )
int given;
{
   int givenfirst, givenlast;
   int other;
   int otherfirst, otherlast;
   int lit1, lit2;
   int c;
   int from1, from2;    /* From equality from1 = from2.  */
   int eq;              /* Type of equality used. Either */
                        /* = or ==>.                     */
   int into, intoend;   /* The into term;                */

   bin_clause = given;
   givenfirst = clauses [ bin_clause + CLAUSE_MAXLITS ];
   givenlast = clauses [ bin_clause + CLAUSE_END ];
 
   for( other = listinuse; other != EMPTYCLAUSELIST; 
        other = clauses [ other + CLAUSE_NEXTINLIST ] )
   {
#if 0
      printstring( &outputfile, "\n========================================\n");
      printstring( &outputfile, "givenclause is ");
      printclause( &outputfile, bin_clause );
      printstring( &outputfile, "otherclause is ");
      printclause( &outputfile, other );
      printstring( &outputfile, "\n");
#endif
      otherfirst = clauses [ other + CLAUSE_MAXLITS ];
      otherlast = clauses [ other + CLAUSE_END ]; 

      for( bin_literal = givenfirst; bin_literal < givenlast;
           bin_literal ++ )
      {
         for( c = otherfirst; c < otherlast; c ++ )
         {
            lit1 = clauses [ bin_literal ];
            lit2 = clauses [c];

            if( termspace [ lit1 ] == OP_EQUAL || 
                termspace [ lit1 ] == OP_REWRITES )
            {
#if 0
               printstring( &outputfile, "Rewriting from given: " );
               printterm( &outputfile, lit1 );
               printstring( &outputfile, "\n" );
#endif
               /* From given into other. */

               eq = termspace [ lit1 ];
               from1 = lit1 + 1;
               from2 = termends [ from1 ];

               if( termspace [ lit2 ] == OP_NOT ) 
                  into = lit2 + 1;
               else
                  into = lit2; 
               
               if( termspace [ into ] == OP_REWRITES ) 
               {
                  into ++ ;
                  intoend = termends [ into ];
               }
               else
               {
                  intoend = termends [ into ];
                  if( operatornoparamod [ termspace [ into ]] )
                     into = intoend;
                  else 
                     into ++ ;
               }

               while( into < intoend )
               {
                  if( termspace [ into ] >= 0 )
                  {
#if 0
                     printstring( &outputfile, "into_other: ");
                     printterm( &outputfile, into );
                     printstring( &outputfile, "  " );
#endif
                     freesubst = 0;
                     if( unify( from1, 0, into, 1 ))
                     {
                        bin_term = from1;  
                        makeparafrom( other, c, into );
                     }
       
                     freesubst = 0; 
                     if( eq == OP_EQUAL && unify( from2, 0, into, 1 ))
                     {
                        bin_term = from2;
                        makeparafrom( other, c, into );
                     }
                  }
#if 1
                  if( termspace [ into ] >= 0 && 
                      operatornoparamod [ termspace [ into ] ] )
                  {
                     into = termends [ into ];
                  }
                  else
                     into ++ ;
#else
                  into ++ ;
#endif
               }
            }

            if( termspace [ lit2 ] == OP_EQUAL ||
                termspace [ lit2 ] == OP_REWRITES ) 
            {
#if 0
               printstring( &outputfile, "Rewriting from other: " );
               printterm( &outputfile, lit2 );
               printstring( &outputfile, "\n" );
#endif
               /* From other into given.   */

               eq = termspace [ lit2 ];
               from1 = lit2 + 1;
               from2 = termends [ from1 ];

               if( termspace [ lit1 ] == OP_NOT ) 
                  into = lit1 + 1;
               else
                  into = lit1;

               if( termspace [ into ] == OP_REWRITES )
               {
                  into ++ ;
                  intoend = termends [ into ];
               }
               else
               {
                  intoend = termends [ into ];
                  if( operatornoparamod [ termspace [ into ]] )
                     into = intoend;
                  else
                     into ++ ;
               }
               while( into < intoend )
               {
                  if( termspace [ into ] >= 0 )
                  {
#if 0
                     printstring( &outputfile, "into_given: ");
                     printterm( &outputfile, into );
                     printstring( &outputfile, "   " );
#endif
                     freesubst = 0;
                     if( unify( from1, 0, into, 1 ))
                     {
                        bin_term = into;
                        makeparainto( other, c, from1 );
                     }

                     freesubst = 0;
                     if( eq == OP_EQUAL && unify( from2, 0, into, 1 ))
                     {
                        bin_term = into;
                        makeparainto( other, c, from2 );
                     }
                  }
#if 1
                  if( termspace [ into ] >= 0 && 
                      operatornoparamod [ termspace [ into ] ] )
                  {
                     into = termends [ into ];
                  }
                  else
                     into ++ ;
#else
                  into ++ ;
#endif
               }
            }
         }
      }
   }
}



/* select the lightest clause from listclauses. */

static int selectlightest( )
{
   int *best, *p;
   int weight, depth, length, nrvars;
   int selected;

   best = &listclauses;
   weight = clauses [ *best + CLAUSE_WEIGHT ];
   depth = clauses [ *best + CLAUSE_DEPTH ];
   length = clauses [ *best + CLAUSE_LENGTH ];
   nrvars = clauses [ *best + CLAUSE_NRVARS ];

   for( p = &clauses [ listclauses + CLAUSE_NEXTINLIST ];
        *p != EMPTYCLAUSELIST;
        p = &clauses [ *p + CLAUSE_NEXTINLIST ] )
   {
      if( clauses [ *p + CLAUSE_WEIGHT ] < weight )
         goto selectp; 
      if( clauses [ *p + CLAUSE_WEIGHT ] == weight )
      {
         if( clauses [ *p + CLAUSE_DEPTH ] < depth )
            goto selectp;
         if( clauses [ *p + CLAUSE_DEPTH ] == depth )
         {
            if( clauses [ *p + CLAUSE_LENGTH ] < length )
               goto selectp;
            if( clauses [ *p + CLAUSE_LENGTH ] == length )
            {
               if( clauses [ *p + CLAUSE_NRVARS ] > nrvars )
                  goto selectp;
            }
         }
      }
      goto next;
   selectp:
      best = p;
      weight = clauses [ *best + CLAUSE_WEIGHT ];
      depth = clauses [ *best + CLAUSE_DEPTH ];
      length = clauses [ *best + CLAUSE_LENGTH ];
      nrvars = clauses [ *best + CLAUSE_NRVARS ]; 
   next:
      ;
   }
   selected = *best;
   *best = clauses [ *best + CLAUSE_NEXTINLIST ];
   return selected;
}

/* select the heaviest clause from listclauses. */

static int selectheaviest( )
{
   int *worst, *p;
   int weight, depth, length, nrvars;
   int selected;

   worst = &listclauses;
   weight = clauses [ *worst + CLAUSE_WEIGHT ];
   depth = clauses [ *worst + CLAUSE_DEPTH ];
   length = clauses [ *worst + CLAUSE_LENGTH ];
   nrvars = clauses [ *worst + CLAUSE_NRVARS ];

   for( p = &clauses [ listclauses + CLAUSE_NEXTINLIST ];
        *p != EMPTYCLAUSELIST;
        p = &clauses [ *p + CLAUSE_NEXTINLIST ] )
   {
      if( clauses [ *p + CLAUSE_WEIGHT ] > weight )
         goto selectp; 
      if( clauses [ *p + CLAUSE_WEIGHT ] == weight )
      {
         if( clauses [ *p + CLAUSE_DEPTH ] > depth )
            goto selectp;
         if( clauses [ *p + CLAUSE_DEPTH ] == depth )
         {
            if( clauses [ *p + CLAUSE_LENGTH ] > length )
               goto selectp;
            if( clauses [ *p + CLAUSE_LENGTH ] == length )
            {
               if( clauses [ *p + CLAUSE_NRVARS ] < nrvars )
                  goto selectp;
            }
         }
      }
      goto next;
   selectp:
      worst = p;
      weight = clauses [ *worst + CLAUSE_WEIGHT ];
      depth = clauses [ *worst + CLAUSE_DEPTH ];
      length = clauses [ *worst + CLAUSE_LENGTH ];
      nrvars = clauses [ *worst + CLAUSE_NRVARS ]; 
   next:
      ;
   }
   selected = *worst;
   *worst = clauses [ *worst + CLAUSE_NEXTINLIST ];
   return selected;
}


/* select the clause with the smallest number from listclauses.   */

static int selectnr( )
{
   int *best, *p;
   int number;
   int selected;

   best = &listclauses;
   number = clauses [ *best + CLAUSE_NUMBER ];

   for( p = &clauses [ listclauses + CLAUSE_NEXTINLIST ];
        *p != EMPTYCLAUSELIST;
        p = &clauses [ *p + CLAUSE_NEXTINLIST ] )
   {
      if( clauses [ *p + CLAUSE_NUMBER ] < number )
         goto selectp; 
      goto next;
   selectp:
      best = p;
      number = clauses [ *best + CLAUSE_NUMBER ];
   next:
      ;
   }
   selected = *best;
   *best = clauses [ *best + CLAUSE_NEXTINLIST ];
   return selected;
}


/* add a clause to proof list, if it has not been added already:  */
/* return 1 if something was added.                               */
/* Clauses are sorted by number.                                  */


static int addtolistproof( clause )
int clause;
{
   
   int *p;
   int nr1, nr2;

   p = &listproof;

   nr1 = clauses [ clause + CLAUSE_NUMBER ];
loop:
   if( *p == EMPTYCLAUSELIST ) goto addhere; 

   if( *p == clause )
      return 0;

   nr2 = clauses [ *p + CLAUSE_NUMBER ];

   if( nr1 < nr2 ) 
      goto addhere;

   p = &clauses [ *p + CLAUSE_NEXTINLIST ];
   goto loop; 

addhere:
   clauses [ clause + CLAUSE_NEXTINLIST ] = *p;
   *p = clause;
   return 1;
}


/* Do a resimplification on the clauses in *p. This includes    */
/* backward demodulation, backward resolution simplification,   */
/* and backward subsumption.                                    */

 
static void resimpclauselist( p, frominuse )
int *p;
int frominuse;
{
   if( showresimp )
   {
      if( frominuse )
         printstring( &outputfile, "Resimplifying inuse:\n" );
      else
         printstring( &outputfile, "Resimplifying clauses:\n" );
   }
 
   while( *p != EMPTYCLAUSELIST ) 
   {
      if( ! resimpclause( p, frominuse ))
      {
         p = &clauses [ *p + CLAUSE_NEXTINLIST ];
      }
   } 
   if( showresimp )
   {
      printstring( &outputfile, "Done\n\n");
   }
}


/* mainloop does the main derivation. It is called with trans, which is */
/* a reference into translations.                                       */
/* If noderivation = 1, only the initial clauses are made.              */
/* This is used by the automode to see the clauses, so that it can      */
/* select the strategy. In that case the clauses are stored in          */
/* listclauses.                                                         */


int mainloop( trans, noderivation )
int trans;
int noderivation;
{
   int i;
   int given;
   int change;

   if( maxlength >= MAXCLAUSELEN - 3 || 
       maxnrvars >= MAXNRVARS - 3 )
   {
      errorwithoutline( "incorrect settings" );
   }

   /* Initialize derivation variables:    */

   resimpinusecnt = 0;
   resimpclausescnt = 0;
   selectoldestcnt = 0;
   showstatuscnt = 0;

   nrgenerated = 0;
   nrkept = 0;
   nrselected = 0;
   nrdeleted = 0;
   nrdeletedinuse = 0;

   subsentry = 0;
   subspropmatch = 0;
   subslitmatch = 0;
   subsfullsubs = 0;
   
   listinuse = EMPTYCLAUSELIST;
   listpassive = EMPTYCLAUSELIST;
   listclauses = EMPTYCLAUSELIST;
   listunfactored = EMPTYCLAUSELIST;
   listproof = EMPTYCLAUSELIST;
   listdeleted = EMPTYCLAUSELIST;
   listdeletedinuse = EMPTYCLAUSELIST;

   treeforwdemod = DT_NIL;
   treeforwrespos = DT_NIL; 
   treeforwresneg = DT_NIL;
   treeforwsubs = DT_NIL;

   incomplete = 0;   /* Wishful thinking. */

   if( setjmp( emptyclause ))
      goto prooffound;

   for( given = translations [ trans ]. startofresult; 
        given < translations [ trans ]. endofresult;
        given = termends [ given ] ) 
   {
      makeinitialclause( given );
   }

   if( noderivation )
   {
      listclauses = listunfactored;
      listunfactored = EMPTYCLAUSELIST;
      return 0;
   }

   listinuse = listunfactored;
   listunfactored = EMPTYCLAUSELIST;

   while( listinuse != EMPTYCLAUSELIST && nrgoals > 0 )
   {
      given = listinuse;
      listinuse = clauses [ listinuse + CLAUSE_NEXTINLIST ];

      clauses [ given + CLAUSE_NEXTINLIST ] = listunfactored;
      listunfactored = given;
         
      nrgoals -- ; 
   }

   for( given = listinuse; given != EMPTYCLAUSELIST; 
        given = clauses [ given + CLAUSE_NEXTINLIST ] )
   {
      processaxiomclause( given );
    
      if( showkept )
      {
         printstring( &outputfile, "Made Axiom: ");
         printclause( &outputfile, given );
      }
   }
   
factors:
   /* If we have too many clauses, then delete a few.           */

   while( listclauses != EMPTYCLAUSELIST && nrkept - nrdeleted > maxclauses )
   {
      incomplete = 1;
      given = selectheaviest( );
      clauses [ given + CLAUSE_NEXTINLIST ] = EMPTYCLAUSELIST;
      deleteclause( &given, 0, DEL_TOOMANY ); 
      goto factors;
   }  

   if( nrselected > maxselected )
   {
      incomplete = 1;
      goto saturationfound;
   }
 
   /* apply unary rules:    */

   while( listunfactored != EMPTYCLAUSELIST )
   {
      given = listunfactored;
      listunfactored = clauses [ listunfactored + CLAUSE_NEXTINLIST ];

      if( usefactor )
         makefactors( given );

      if( useeqrefl )
         makeeqrefls( given );

      if( useeqfact )
         makeeqfacts( given );

      /* add given to listclauses: */

      clauses [ given + CLAUSE_NEXTINLIST ] = listclauses;
      listclauses = given;
   }



resolution:
   if( showstatuscnt >= showstatus )
   {
      showstatuscnt = 0;

      printstring( &outputfile, "\n");

      printstring( &outputfile, "Intermediate Status:\n");

      printstring( &outputfile, "Generated:    " );
      printint( &outputfile, nrgenerated );
      printstring( &outputfile, "\n" );

      printstring( &outputfile, "Kept:         " );
      printint( &outputfile, nrkept );
      printstring( &outputfile, "\n" );

      printstring( &outputfile, "Inuse:        " );
      printint( &outputfile, nrselected );
      printstring( &outputfile, "\n");

      printstring( &outputfile, "Deleted:      " );
      printint( &outputfile, nrdeleted );
      printstring( &outputfile, "\n" );

      printstring( &outputfile, "Deletedinuse: " );
      printint( &outputfile, nrdeletedinuse );
      printstring( &outputfile, "\n" );

      printstring( &outputfile, "\n" );
   }
   
   if( listclauses == EMPTYCLAUSELIST )
   {
      resimpclauselist( &listinuse, 1 );
      if( listunfactored != EMPTYCLAUSELIST ) 
         goto factors;

      if( listunfactored != EMPTYCLAUSELIST || 
          listclauses != EMPTYCLAUSELIST ) 
      { 
         errorwithoutline( "internal problem" );
      }
      goto saturationfound;
   }

   if( resimpinusecnt >= resimpinuse  )
   {
      resimpclauselist( &listinuse, 1 );
      resimpinusecnt = 0;
      goto factors;
   }

   if( resimpclausescnt >= resimpclauses )
   {
      resimpclauselist( &listclauses, 0 );
      resimpclausescnt = 0;
      goto factors;
   } 

   /* select the most preferred clause: */

   if( selectoldestcnt >= selectoldest )
   {
      selectoldestcnt = 0;
      given = selectnr( );
   }
   else
      given = selectlightest( );

   clauses [ given + CLAUSE_NEXTINLIST ] = EMPTYCLAUSELIST;
   selectoldestcnt ++ ;

   if( resimpclause( &given, 0 ))
      goto factors;

   nrselected ++ ; 

   /* show the clause if the option is set. */

   if( showselected )
   {
      printstring( &outputfile, "Select [" );
      if( !showkept )
      {
         printint( &outputfile, nrkept );
            printstring( &outputfile, ", ");
      }
      printint( &outputfile, nrselected );
      printstring( &outputfile, "]: " ); 
      printclause( &outputfile, given );
   }

   /* add it to listinuse. */

   clauses [ given + CLAUSE_NEXTINLIST ] = listinuse;
   listinuse = given; 

   if( useres ) 
      searchbinary( given );

   if( useparamod )
      searchparamod( given );

   goto factors; 

saturationfound:
   /* A saturation is found if increasemaxweight is off, or otherwise no */
   /* clause has been discarded because of the weight restriction.       */

   if( incomplete == 0 ) 
   {
      printstring( &outputfile, "\n\n" ); 
      printstring( &outputfile, "found a saturation!\n\n" );
  
      printclauselist( &outputfile, listinuse ); 
      printclauselist( &outputfile, listclauses );
      printclauselist( &outputfile, listunfactored );

      printstring( &outputfile, "\n\n" );
      printstring( &outputfile, "found a saturation!\n\n" ); 
      return 1; 
   }
   else
   {
      /* Ceci n'est pas une saturation!   */

      printstring( &outputfile, "Failed to find proof!\n");
      printstring( &outputfile, "maxweight =   " );
      printint( &outputfile, maxweight );
      printstring( &outputfile, "\n" );
      if( maxclauses > 0 )
      {
         printstring( &outputfile, "maxclauses = " );
         printint( &outputfile, maxclauses );
         printstring( &outputfile, "\n" );
      }
      printstring( &outputfile, "Generated: " );
      printint( &outputfile, nrgenerated );
      printstring( &outputfile, "\n");
      printstring( &outputfile, "Kept: ");
      printint( &outputfile, nrkept );
      printstring( &outputfile, "\n\n\n" );
      return 0;
   }

prooffound:
#if 1
   /* collect the clauses that were used. */

   printstring( &outputfile, "\n" );
   printstring( &outputfile, "Bliksems!, er is een bewijs:\n\n" );

   for( given = listunfactored; given != EMPTYCLAUSELIST; 
        given = clauses [ given + CLAUSE_NEXTINLIST ] )
   {
      if( clauses [ given + CLAUSE_LENGTH ] == 0 ) 
      {
         (void) addtolistproof( given ); 
      }
   }
   change = 1;
   while( change )
   {
      change = 0;
      for( given = listproof; given != EMPTYCLAUSELIST; 
           given = clauses [ given + CLAUSE_NEXTINLIST ] ) 
      {
         for( i = clauses [ given + CLAUSE_DERIVATION ]; 
              i < clauses [ given + CLAUSE_LITERALS ]; 
              i ++ ) 
         {
            if( clauses [i] >= 0 )
               change = change | addtolistproof( clauses [i] );
         }
      }
   }
   printclauselist( &outputfile, listproof );
   printstring( &outputfile, "\n\n" );
   printstring( &outputfile, "found a proof!\n\n" ); 
#endif

   if( totalproof )
   {
      checkproof( translations [ trans ]. startofresult, 
                  translations [ trans ]. endofresult, 
                  listproof );
   }

#if 0
   while( listinuse != EMPTYCLAUSELIST )
   {
      given = listinuse;
      listinuse = clauses [ listinuse + CLAUSE_NEXTINLIST ];
      processdeletedclause( given, 1, DEL_SIMP );
   }
   while( listproof != EMPTYCLAUSELIST ) 
   {
      given = listproof;
      listproof = clauses [ listproof + CLAUSE_NEXTINLIST ];
      processdeletedclause( given, 1, DEL_BSUBS );
   }
   while( listclauses != EMPTYCLAUSELIST ) 
   {
      given = listclauses;
      listclauses = clauses [ listclauses + CLAUSE_NEXTINLIST ];
      processdeletedclause( given, 0, DEL_BSUBS );
   }
#endif
 
#if 0
   printstring( &outputfile, "Statistics on Node Visits:\n\n");
   printstats( );
   printstring( &outputfile, "\n" );

   printstring( &outputfile, "Forwsubs:\n\n" ); 
   counttree( treeforwsubs ); 
   printstring( &outputfile, "\n" );

   printstring( &outputfile, "Forwdemod:\n\n" );
   counttree( treeforwdemod );
   printstring( &outputfile, "\n" );

   printstring( &outputfile, "Forwrespos:\n\n" ); 
   counttree( treeforwrespos ); 
   printstring( &outputfile, "\n" );

   printstring( &outputfile, "Forwresneg:\n\n" );
   counttree( treeforwresneg );
   printstring( &outputfile, "\n" );

#endif
   return 1;
}


