
/****************************************************************/
/* file simplify.c contains simplification rules.               */
/* Bliksem 1.10                                                 */
/****************************************************************/

#define CLASS extern
#include "basics.h"
#include "extfunc.h"
#include "resolution.h"



/* make fresh copies of literals [], with the exception of           */
/* literals [ toskip ].                                              */

static void copyexcept( toskip )
int toskip;
{
   int i1, i2;
   int firstterm;

   i2 = 0;
   firstterm = freetermspace;
   for( i1 = 0; i1 < nrliterals; i1 ++ )
   {
      if( i1 != toskip )
         literals [ i2 ++ ] = copyterm( literals [i1] );
   }
   nrliterals = i2;
   maketermends( firstterm, freetermspace );
}


/* move literals [ 0 .. nrliterals ] from from to to.               */

static void moveliterals( from, to )
int from;
int to;
{
   int diff;
   int i;
   int firstterm;
 
   if( to > from ) 
   {
      internalproblem( "moving problem!");
      exit(0);
   }

   diff = from - to;
   firstterm = to;

   while( from < freetermspace )
      termspace [ to ++ ] = termspace [ from ++ ];
   freetermspace = to;
   for( i = 0; i < nrliterals; i ++ )
      literals [i] -= diff;
   maketermends( firstterm, freetermspace ); 
}



/**************************************************************/
/* Do factoring and eqrefl simplification.                    */
/* This means removing identical literals and removing        */
/* literals of the form ! t = t.                              */
/* Question: Does there exist an eqrefl-based notion of       */
/* condensing?                                                */
/**************************************************************/


void simpfacteqrefl( firstterm )
int firstterm;
{
   int i1, i2;
   int lit1, lit2;
   int firstterm2;

restart1:
   if( useeqrefl )
   {
      for( i1 = 0; i1 < nrliterals; i1 ++ )
      {
         lit1 = literals [ i1 ];
         if( termspace [ lit1 ] == OP_NOT ) 
         {
            lit1 ++ ;
            if( termspace [ lit1 ] == OP_EQUAL || 
                termspace [ lit1 ] == OP_REWRITES ) 
            {
               if( termequal( lit1 + 1, termends [ lit1 + 1 ] ))
               {
                  firstterm2 = freetermspace;
                  copyexcept( i1 );
                  moveliterals( firstterm2, firstterm );
 
                  if( freeparents < MAXPARENTS - 5 ) 
                     parents [ freeparents ++ ] = SIMP_EQREFL;
                  goto restart1;
               } 
            }
         }
      }                  
   }

restart2:
   if( usefactor )
   {
      for( i1 = 0; i1 < nrliterals; i1 ++ )
      {
         lit1 = literals [ i1 ];
         for( i2 = 0; i2 < i1; i2 ++ )
         {
            lit2 = literals [ i2 ];
            if( termequal( lit1, lit2 ))
            {
               firstterm2 = freetermspace;
               copyexcept( i1 );
               moveliterals( firstterm2, firstterm );

               if( freeparents < MAXPARENTS - 5 ) 
                  parents [ freeparents ++ ] = SIMP_FACTOR;
               goto restart2;
            }
         }
      }
   }
}       



/* Return 1 if clause can be matched into literals [], in the       */
/* context of matching []. matching [] has been constructed by      */
/* matching fromlit into intolit. Therefore fromlit should not be   */
/* matched, and no literal should be matched into                   */
/* literals [ intolit ].                                            */
/* Since currently clauses [ fromlit ] should contain all           */
/* variables in the clause, no backtracking is necessary.           */
/* ***** This should be improved, at some beautiful day.            */


static int lit_subsumes( clause, fromlit, intolit )
int clause;
int fromlit;
int intolit;
{
   int clausestart, clauseend;
   int i, j;
   int lit1, lit2;
   int at1, at2;
   int freematching2;

   freematching2 = freematching;
   clausestart = clauses [ clause + CLAUSE_LITERALS ];
   clauseend = clauses [ clause + CLAUSE_END ];

   for( i = clausestart; i < clauseend; i ++ )
   {
      if( i != fromlit )
      {
         lit1 = clauses [i];
         for( j = 0; j < nrliterals; j ++ )
         {
            if( j != intolit )
            {
               lit2 = literals [j];
               if( match( lit1, lit2 ))
                  goto success;
               if( substype == SUBS_EQREWR ) 
               {
                  if( termspace [ lit1 ] == OP_NOT &&
                      termspace [ lit2 ] == OP_NOT ) 
                  {
                     at1 = lit1 + 1;
                     at2 = lit2 + 1;
                  }
                  else
                  {
                     at1 = lit1;
                     at2 = lit2;
                  }
                  if( ( termspace [ at1 ] == OP_EQUAL ||
                        termspace [ at1 ] == OP_REWRITES ) &&
                      ( termspace [ at2 ] == OP_EQUAL || 
                        termspace [ at2 ] == OP_REWRITES ) )
                  {
                     if( match( at1 + 1, at2 + 1 ) &&
                         match( termends [ at1 + 1 ], termends [ at2 + 1 ] ))
                        goto success;
                     if( match( at1 + 1, termends [ at2 + 1 ] ) &&
                         match( termends [ at1 + 1 ], at2 + 1 ))
                        goto success;
                  }
               }
            }
         }
         freematching = freematching2;
         return 0;
      } 
   success:
      ;
   }
   return 1;
}    



/********************************************************************/
/*               Demodulation.                                      */
/********************************************************************/


static int dem_forward;     /* 1 if it is a forward demodulation.   */
static int dem_firstterm;   /* The first term, if dem_forward = 1.  */
static int dem_clause;      /* Used only when dem_forward = 0,      */
                            /* to check that a clause does not      */
                            /* rewrite itself.                      */
static int dem_lit;         /* Reference into literals [].          */
static int dem_term;        /* The term, in termspace [].           */
static jmp_buf dem_rewritten;  /* jumped to after a rewrite.        */
static int dem_nr;          /* How often rewritten.                 */


/* Delete or add the demodulators that can be obtained from clause    */
/* to the discrimination tree treeforwdemod.                          */


void dt_demod( clause, adding ) 
{
   int i;
   int nr;
   int lit, term;

   if( clauses [ clause + CLAUSE_LENGTH ] <= usesimpdemod )
   {
      nr = clauses [ clause + CLAUSE_NRVARS ];
      for( i = clauses [ clause + CLAUSE_LITERALS ];
           i < clauses [ clause + CLAUSE_END ];
           i ++ )
      {
         lit = clauses [i];
         if( termspace [ lit ] == OP_REWRITES ) 
         {
            term = lit + 1;
            if( termnrvars( term ) == nr ) 
            {
               if( adding )
                  addtotree( &treeforwdemod, clause, i, term );
               else
                  deletefromtree( &treeforwdemod, clause, i, term );
            }
         }
         if( termspace [ lit ] == OP_EQUAL && demodaposteriori )
         {
            term = lit + 1;
            if( termnrvars( term ) == nr )
            {
               if( adding )
                  addtotree( &treeforwdemod, clause, i, term );
               else
                  deletefromtree( &treeforwdemod, clause, i, term );
            }

            term = termends [ lit + 1 ];
            if( termnrvars( term ) == nr )
            {
               if( adding )
                  addtotree( &treeforwdemod, clause, i, term );
               else
                  deletefromtree( &treeforwdemod, clause, i, term );
            }
         }
      }
   }
}


/* Replace t1 by t2 everywhere in term, at replacable positions.    */
/* return the result.                                               */

static int replace( literal, t1, t2 )
int literal;
int t1;
int t2;
{
   int res;
   int literalend;

   res = freetermspace;
   literalend = termends [ literal ];

   if( termspace [ literal ] == OP_NOT ) 
      termspace [ freetermspace ++ ] = termspace [ literal ++ ];
   termspace [ freetermspace ++ ] = termspace [ literal ++ ];
   while( literal < literalend )
   {
      if( termequal( literal, t1 ))
      {
         (void) copyterm( t2 );   
         literal = termends [ literal ];
      }
      else
         termspace [ freetermspace ++ ] = termspace [ literal ++ ];
      if( freetermspace >= MAXTERMSPACE - 10 ) 
         errorwithoutline( "termspace is full" );
   }
   return res;
}
 



/* Retrieved a potential demodulator from treeforwdemod.    */

static void demod2( clause, literal, t1 )
int clause;
int literal;
int t1;
{
   int t2;
   int t1s, t2s;
   int m;
   int i,j;

   /* Like with subsumption, the from clause should be not longer, */
   /* than the into clause.                                        */
   /* Otherwise factors could be deleted.                          */

   if( clauses [ clause + CLAUSE_LENGTH ] > nrliterals ) 
      return;
   if( dem_clause == clause ) 
      return;              /* A clause should not rewrite itself.  */
 
   /* Check the side literals in the context of the matching.  */

   if( ! lit_subsumes( clause, literal, dem_lit ))
      return;

   /* Produce t2.                                   */

   if( termspace [ clauses [ literal ] ] != OP_EQUAL && 
       termspace [ clauses [ literal ] ] != OP_REWRITES ) 
   {
      internalproblem( "internal problem!");
      exit(0);
   } 

   if( t1 == clauses [ literal ] + 1 )
      t2 = termends [ t1 ];
   else
      t2 = clauses [ literal ] + 1;

   /* produce the terms after the matching.   */

   t1s = matchsubst( t1 );
   t2s = matchsubst( t2 );
   maketermends( t1s, freetermspace );

   /* Do the aposteriori check if demodaposteriori == 1, and the */
   /* equality was not directed already.                         */

   if( demodaposteriori &&
       termspace [ clauses [ literal ]] == OP_EQUAL &&
       termcompare( t1s, t2s ) != COMP_MORE ) 
   {
      freetermspace = t1s;
      return;
   }

#if 0
   if( termspace [ clauses [ literal ]] == OP_EQUAL )
   {
      printf("danger!\n");
   }
 
   if( dem_forward )
      printstring( &outputfile, "Found Forward Demodulator:\n" );
   else
      printstring( &outputfile, "Found Backward Demodulator:\n");
   printclause( &outputfile, clause );
   printterm( &outputfile, t1 );
   printstring( &outputfile, " ==> " );
   printterm( &outputfile, t2 );
   printstring( &outputfile, "\n" );
   printmatching( &outputfile );
#endif

   /* Produce the rewritten literal. */

   m = replace( literals [ dem_lit ], t1s, t2s );
   i = t1s;
   j = m;
   while( j < freetermspace )
      termspace [ i ++ ] = termspace [ j ++ ];
   freetermspace = i;
   maketermends( t1s, freetermspace );

   /* Produce the side literals:     */

   copyexcept( dem_lit );

   /* And add the rewritten literal back:   */

   literals [ nrliterals ++ ] = t1s;

   if( freeparents <= MAXPARENTS - 5 ) 
   {
      parents [ freeparents ++ ] = SIMP_DEMOD;
      parents [ freeparents ++ ] = clause;
   }

   if( dem_forward )
      moveliterals( t1s, dem_firstterm );

#if 0
   printstring( &outputfile, "Result: { ");
   for( dem_lit = 0; dem_lit < nrliterals; dem_lit ++ )
   {
      printterm( &outputfile, literals [ dem_lit ] );
      if( dem_lit < nrliterals - 1 )
         printstring( &outputfile, ", ");
   }
   printstring( &outputfile, " }\n\n");
#endif

   dem_nr ++ ;
   longjmp( dem_rewritten, 1 );
}

 

/* Try to simplify literals [ 0 .. nrliterals ] by demodulation. In the */
/* case that forward = 1, we are doing forward demodulation. In that    */
/* case termspace [ firstterm .. freetermspace ] contains the terms,    */
/* and they are overwritten. If forward = 0, we are doing backward      */
/* demodulation. In this case clause is used to check that we are not   */
/* rewriting a clause with itself, and no overwriting will take place.  */
/* Instead new terms are generated for literals [ 0 .. nrliterals ].    */
/* They should be processed as a newly generated clause.                */
/* simpdemod returns 1 if at least one rewrite was made.                */

int simpdemod( firstterm, clause, forward )
int firstterm;
int clause;
int forward;
{
   int litend;

   if( forward == 1 )
   {
      dem_firstterm = firstterm;
      dem_clause = -1;
   }
   else
   {
      dem_firstterm = -1;
      dem_clause = clause;
   }
   dem_forward = forward;
   dem_nr = 0;   

   if( setjmp( dem_rewritten ))
   {
      if( dem_forward == 0 )
         return 1;
   }

   for( dem_lit = 0; dem_lit < nrliterals; dem_lit ++ )
   {
      dem_term = literals [ dem_lit ];
      litend = termends [ dem_term ];

      if( termspace [ dem_term ] == OP_NOT ) 
         dem_term ++ ;
      dem_term ++ ;
      while( dem_term < litend )
      {
         retrievefromtree( treeforwdemod, dem_term, demod2 );
         dem_term ++ ;
      }
   }
   return dem_nr;
}
   




/******************************************************/
/* return 1 if literals [] contains a literal t = t,  */ 
/* or a complementary pair of literals.               */
/******************************************************/

int literalsistautology( )
{
   int i1, i2;
   int lit1, lit2;

   for( i1 = 0; i1 < nrliterals; i1 ++ )
   {
      /* Equality tautology?          */

      lit1 = literals [ i1 ];
      if( termspace [ lit1 ] == OP_EQUAL ||
          termspace [ lit1 ] == OP_REWRITES ) 
      {
         if( termequal( lit1 + 1, termends [ lit1 + 1 ] ))
            return 1;
      }
      if( termspace [ lit1 ] != OP_NOT ) 
      {
         /* Propositional tautology?   */

         for( i2 = 0; i2 < nrliterals; i2 ++ )
         {
            lit2 = literals [ i2 ];
            if( termspace [ lit2 ] == OP_NOT &&
                termequal( lit1, lit2 + 1 ))
               return 1;
         }
      }
   }
   return 0;
}



/*******************************************************/
/* Do selfreduction on literals [0 .. nrliterals ].    */
/* This means trying to rewrite by the negative        */
/* equalities.                                         */
/*******************************************************/

void selfreduceliterals( firstterm )
int firstterm;
{
   return;
}


/*************************************************************/
/*               Resolution Simplification.                  */
/*************************************************************/

static int res_forward;     /* 1 if we are doing forward resolution  */
                            /* simplification.                       */
static int res_firstterm;   /* Used for forward simplification, s.t. */
                            /* we can replace the old literals.      */
static int res_clause;      /* Used for checking that a clause does  */
                            /* not simplify itself.                  */
static int res_lit;         /* Reference in literals [].             */
static int res_term;        /* Reference in termspace [].            */

static jmp_buf res_rewritten;
static int res_nr;          /* Counts the nr. of rewrites.           */


/* add or delete resolution simplifiers based on clause, either to    */
/* treeforwrespos or treeforwnegpos.                                  */

void dt_simpres( clause, adding )
int clause;
int adding;
{
   int i;
   int nr;
   int lit;

   if( clauses [ clause + CLAUSE_LENGTH ] <= usesimpres )
   {
      nr = clauses [ clause + CLAUSE_NRVARS ];
      for( i = clauses [ clause + CLAUSE_LITERALS ];
           i < clauses [ clause + CLAUSE_END ];
           i ++ )
      {
         if( termnrvars( clauses [i] ) == nr )
         {
            lit = clauses [i];
            if( termspace [ lit ] == OP_NOT ) 
            {
               if( adding )
                  addtotree( &treeforwresneg, clause, i, lit + 1 );
               else
                  deletefromtree( &treeforwresneg, clause, i, lit + 1 );
            }
            else
            {
               if( adding )
                  addtotree( &treeforwrespos, clause, i, lit );
               else
                  deletefromtree( &treeforwrespos, clause, i, lit );
            }
         }
      }   
   }
}



static void ressimp2( clause, literal, term )
int clause;
int literal;
int term;
{
   int base;

   if( clauses [ clause + CLAUSE_LENGTH ] > nrliterals )
      return;
   if( ! lit_subsumes( clause, literal, res_lit ))
      return;

#if 0
   if( res_forward )
      printstring( &outputfile, "Found Forward Resolution Simplifier:\n" );
   else
      printstring( &outputfile, "Found Backward Resolution Simplifier:\n");

   printclause( &outputfile, clause );
   printterm( &outputfile, clauses [ literal ] );
   printstring( &outputfile, "\n" );
   printmatching( &outputfile );
#endif

   base = freetermspace;
   copyexcept( res_lit );

   if( freeparents <= MAXPARENTS - 5 ) 
   {
      parents [ freeparents ++ ] = SIMP_RES;
      parents [ freeparents ++ ] = clause;
   }

   if( res_forward )
      moveliterals( base, res_firstterm );

#if 0
   printstring( &outputfile, "Result: { " );
   for( res_lit = 0; res_lit < nrliterals; res_lit ++ )
   {
      printterm( &outputfile, literals [ res_lit ] );
      if( res_lit < nrliterals - 1 )
         printstring( &outputfile, ", " );
   }
   printstring( &outputfile, " }\n\n" );
#endif

   res_nr ++ ; 
   longjmp( res_rewritten, 1 );
}



/* Try to resolution simplify literals [ 0 .. nrliterals ]. See simpdemod */
/* for details.                                                           */

int simpreso( firstterm, clause, forward )
int firstterm;
int clause;
int forward;
{

   if( forward == 1 )
   {
      res_firstterm = firstterm;
      res_clause = -1;
   }
   else
   {
      res_firstterm = -1;
      res_clause = clause;
   }
   res_forward = forward;
   res_nr = 0;

   if( setjmp( res_rewritten ))
   {
      if( res_forward == 1 )
         return 1;
   } 

   for( res_lit = 0; res_lit < nrliterals; res_lit ++ )
   {
      res_term = literals [ res_lit ];
      if( termspace [ res_term ] == OP_NOT ) 
      {
         res_term ++ ;
         retrievefromtree( treeforwrespos, res_term, ressimp2 );
      }
      else
      {
         retrievefromtree( treeforwresneg, res_term, ressimp2 );
      }  
   }
   return res_nr;
}



